#
#   Copyright (c) 2014-2015 eGauge Systems LLC
# 	4730 Walnut St, Suite 110
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
from __future__ import print_function

import math
import re

from datetime import date, datetime, timedelta
from decimal import Decimal
from django.core.validators import MinValueValidator, MaxValueValidator
from django.core.urlresolvers import reverse
from django.db import models
from django.utils.html import escape
from django.utils import timezone

PART_PRICE_VALIDATORS = [ MinValueValidator(Decimal('0.000001')) ]

def strchoice(choices, val):
    for key, name in choices:
        if key == val:
            return name
    return val

def part_cost(line_cost, qty):
    if qty <= 0:
        return 0
    return Decimal('%.6f' % (float(line_cost) / qty))

def format_part_number(part):
    return 'EP%05u' % part

def html_part_link(part):
    """Return an HTML link for PART.  PART may be either a Part object or
    a part number.
    """
    if isinstance(part, Part):
        pn = part.id
    else:
        pn = part
    name = format_part_number(pn)
    url = reverse('epic:part_detail', kwargs={'pk': pn})
    return '<a href="%s">%s</a>' % (url, name)

def html_list_or_none(list):
    """If list is empty, return 'none', otherwise, join the list using ','
    as a separator.
    """
    html = ', '.join(list)
    if html != '':
        return html
    return '<em>&mdash;None&mdash;</em>'

class Part(models.Model):
    MOUNTING_SMD = 0
    MOUNTING_THD = 1
    MOUNTING_CHASSIS = 2
    MOUNTING_FREE = 3
    MOUNTING_CHOICES = (
        (MOUNTING_SMD,		'SMD'),
        (MOUNTING_THD,		'THD'),
        (MOUNTING_CHASSIS,	'chassis'),
        (MOUNTING_FREE,		'free')
    )
    STATUS_PREVIEW = 0
    STATUS_ACTIVE = 1
    STATUS_DEPRECATED = 2
    STATUS_OBSOLETE = 3
    STATUS_CHOICES = (
        (STATUS_PREVIEW,	'preview'),
        (STATUS_ACTIVE,		'active'),
        (STATUS_DEPRECATED,	'deprecated'),
        (STATUS_OBSOLETE,	'obsolete'),
    )
    # deprecated parts are orderable, but preview or obsolete parts
    # are not:
    STATUS_ORDERABLE = ( STATUS_ACTIVE, STATUS_DEPRECATED )
    # what modified the part's BOM (assembly-items list) last (user or tool)?:
    LAST_MOD_TYPE_USER	= 0
    LAST_MOD_TYPE_TOOL	= 1
    LAST_MOD_CHOICES = (
        (LAST_MOD_TYPE_USER,	'user'),
        (LAST_MOD_TYPE_TOOL,	'tool')
    )

    val = models.CharField(max_length=31, verbose_name='Value', blank=True,
                           default='',
                           help_text='The primary value of the part '
                           'such as resistance for a resistor or '
                           'capacitance for a capacitor.')
    descr = models.CharField(max_length=127, verbose_name='Description',
                             blank=True, default='',
                             help_text='Brief listing of key parameters of '
                             'the part, such as temperature range, drift, '
                             'max. voltages, etc.')
    footprint = models.CharField(max_length=63, blank=True, default='',
                                 help_text='The part\'s PCB footprint.')
    mfg = models.CharField(max_length=31, verbose_name='Manufacturer',
                           help_text='The name of the manufacturer of the '
                           'part.')
    mfg_pn = models.CharField(max_length=31,
                              verbose_name='Manufacturer\'s Part #')
    mounting = models.IntegerField(choices=MOUNTING_CHOICES,
                                   default=MOUNTING_SMD, blank=False,
                                   verbose_name='Mounting Type',
                                   help_text='How the part is mounted to a '
                                   'PCB (free if it\'s not mounted at all, '
                                   'such as a plug).')
    target_price = models.DecimalField(max_digits=10, decimal_places=6,
                                       validators=PART_PRICE_VALIDATORS,
                                       verbose_name='Target price [$]',
                                       help_text='The expected price of the '
                                       'part.')
    overage = models.DecimalField(max_digits=6, decimal_places=3,
                                  validators=[
                                      MinValueValidator(Decimal('0')),
                                      MaxValueValidator(Decimal('100')) ],
                                  verbose_name='Overage [%]',
                                  help_text='Worst-case percentage of parts '
                                  'we expect to lose due to attrition at the '
                                  'assembly-house.')
    spq = models.IntegerField(default=1,
                              verbose_name='Standard-Package Qty',
                              help_text='The number of parts in '
                              'a standard package.  For example, 10,000 '
                              'pieces in a reel.')
    lead_time = models.PositiveIntegerField(verbose_name='Lead-time [weeks]',
                                            help_text='Lead-time in weeks.')
    status = models.IntegerField(choices=STATUS_CHOICES,
                                 default=STATUS_PREVIEW, blank=False,
                                 verbose_name='Life-time Status',
                                 help_text='The life-time status of this '
                                 'part. Parts marked `preview\' and '
                                 '`obsolete\' are not considered orderable.')
    substitute = models.ForeignKey('self', null=True, blank=True,
                                   verbose_name='Substitutes',
                                   help_text='List of other parts that are '
                                   '(identical) substitutes for this part.')
    last_bom_mod_type = models.IntegerField(choices=LAST_MOD_CHOICES,
                                            default=LAST_MOD_TYPE_USER,
                                            blank=False)
    last_bom_mod_name = models.CharField(max_length=31, blank=True, default='',
                                         help_text='Name of entity '
                                         'which last modified this part.')

    equiv_parts = None
    best_parts = {}

    class Meta:
        unique_together = ('mfg', 'mfg_pn')
        index_together = [
            [ 'mfg', 'mfg_pn' ]
        ]

    def __str__(self):
        return '%s' % (format_part_number(self.id))

    @classmethod
    def flush_caches(cls, sender, **kwargs):
        """See signals.flush_caches()."""
        # clear the caches when a part changes:
        Part.equiv_parts = None
        Part.best_parts = {}

    def equivalent_parts(self):
        """Return the set of equivalent parts for this part.  The set
        always includes the part itself, so it'll never be empty"""

        def get_set(part):
            if part.id not in Part.equiv_parts:
                Part.equiv_parts[part.id] = set([ part ])
            return Part.equiv_parts[part.id]

        if self.id is None:
            # newly created parts don't have a valid id yet so they can't
            # be hashed, which also means nobody can be referring to it,
            # which means its equivalent-parts-set is empty.
            return set([ ])

        if Part.equiv_parts is None:
            Part.equiv_parts = {}
            for p in Part.objects.exclude(substitute__isnull=True):
                union  = get_set(p)
                union |= get_set(p.substitute)
                for q in union:
                    Part.equiv_parts[q.id] = union
        return set(get_set(self))

    def best_part(part):
        """Among the set of parts equivalent to this part, find orderable or
        lowest-cost part.  Returns part itself if nothing better is
        available.

        """
        if part.id in Part.best_parts:
            return Part.best_parts[part.id]

        best_part = part		# may not be orderable...
        for p in part.equivalent_parts():
            if p == part or not p.is_orderable():
                continue
            target_pr = p.target_price
            best_pr = best_part.target_price
            if not best_part.is_orderable() \
               or target_pr < best_pr \
               or (target_pr == best_pr and p.id < best_part.id):
                best_part = p
        for p in part.equivalent_parts():
            Part.best_parts[part.id] = best_part
        return best_part

    def best_vendor(part):
        """Return the vendor that has the lowest price for this part."""
        vps = Vendor_Part.objects.filter(part_id=part.id) \
                                 .order_by('price')
        for vp in vps:
            if vp.status in Part.STATUS_ORDERABLE:
                return vp.vendor
        return None

    def is_orderable(self):
        return self.status in Part.STATUS_ORDERABLE

    def assembly_items(self):
        return Assembly_Item.objects.filter(assy_id=self.id)

    def html_link(self):
        return html_part_link(self)

    def avg_cost(self):
        """Get average cost for a part, including any of its substitutes.  The
        average is calculated across orders with shipments within the past
        365 days.  If there is no shipment within 365, the cost is
        calculated based on the most recent shipment alone.  If there are
        no shipments at all, return the part's target price.
        """
        total_cost = 0
        total_qty = 0
        cut_off_ts = timezone.now() - timedelta(days=365)

        part_id_list = [ p.id for p in self.equivalent_parts() ]
        ship_id_list = Shipment.objects.all() \
                                       .values_list('transaction_ptr_id',
                                                     flat=True)
        deltas = Delta.objects.filter(adj__gt=0) \
                              .filter(part_id__in=part_id_list) \
                              .filter(txtn_id__in=ship_id_list) \
                              .order_by('txtn_id')

        for d in deltas:
            cost = d.txtn.shipment.cost_details()
            overhead_ratio = 0
            if cost['parts'] > 0:
                overhead_ratio = cost['overhead'] / cost['parts']

            for p_id in part_id_list:
                if p_id not in cost['part_detail']:
                    continue
                pc = cost['part_detail'][p_id]
                if d.txtn.shipment.from_warehouse is None:
                    piece_cost = float(pc['cost']) / pc['qty']
                    total_cost += (d.adj * piece_cost) \
                                  * (1 + float(overhead_ratio))
                    total_qty += d.adj
                else:
                    # For inter-warehouse-shipments, allocate the shipping
                    # costs in proportion to the target-cost of the parts
                    part_ratio = (pc['qty'] * d.part.target_price) \
                                 / cost['parts_target']
                    total_cost += float(part_ratio * cost['overhead'])
                if d.txtn.ts <= cut_off_ts:
                    break
        if total_qty == 0:
            return self.target_price
        return part_cost(total_cost, total_qty)

    def strstatus(self):
        return strchoice(Part.STATUS_CHOICES, self.status)

class Vendor(models.Model):
    name = models.CharField(max_length=31, verbose_name='Vendor Name',
                             db_index=True, unique=True)
    search_url = models.CharField(max_length=127,
                                   verbose_name='Search URL Pattern',
                                   default='', blank=True,
                                   help_text='This pattern defines how to '
                                   'search for a particular part on the '
                                   'vendor\'s website.  %(vendor_pn)s gets '
                                   'replaced by the vendor\'s part-number, '
                                   '%(mfg)s by the manufacturer\'s name, and '
                                   '%(mfg_pn)s by the manufacturer\'s '
                                   'part-number.')
    def __str__(self):
        return '%s' % self.name

    def html_link(self):
        url = reverse('epic:vendor_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.name)

    def is_vendor(self):
        return True

    class Meta:
        ordering = [ 'name' ]

class Vendor_Part(models.Model):
    part = models.ForeignKey(Part, verbose_name='Part #')
    vendor = models.ForeignKey(Vendor)
    vendor_pn = models.CharField(max_length=31,
                                 verbose_name='Vendor\'s Part #')
    price = models.DecimalField(max_digits=10, decimal_places=6,
                                validators=PART_PRICE_VALIDATORS)
    status = models.IntegerField(choices=Part.STATUS_CHOICES,
                                 default=Part.STATUS_ACTIVE, blank=False,
                                 verbose_name='Life-time Status',
                                 help_text='The life-time status of this '
                                 'vendor part.')

    class Meta:
        unique_together = (
            ('vendor', 'part'),
            ('vendor', 'vendor_pn')
        )

    def __str__(self):
        return 'Vendor_Part %s: %s %s' % (format_part_number(self.part.id),
                                          self.vendor.name, self.vendor_pn)

    def html_link(self):
        if self.vendor.search_url == '':
            return escape(self.vendor_pn)
        url = self.vendor.search_url % \
              {
                  'mfg'		: escape(self.part.mfg),
                  'mfg_pn'	: escape(self.part.mfg_pn),
                  'vendor_pn'	: escape(self.vendor_pn)
              }
        return '<a href="%s" target="part_lookup">%s</a>' % \
            (url, self.vendor_pn)

    def strstatus(self):
        return strchoice(Part.STATUS_CHOICES, self.status)

    @staticmethod
    def get(part_id, vendor_id):
        result = Vendor_Part.objects \
                            .filter(part_id=part_id) \
                            .filter(vendor_id=vendor_id)
        if len(result) != 1:
            return None
        return result[0]

class Warehouse(models.Model):
    name = models.CharField(max_length=31, unique=True,
                            help_text='The name of the warehouse.')
    address = models.TextField(blank=True, default='', help_text='The '
                               'shipping address for the warehouse.')

    def __str__(self):
        return '%s' % self.name

    def html_link(self):
        url = reverse('epic:warehouse_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.name)

    def inventories(self):
        return Inventory.objects.filter(warehouse_id=self.id)

    @staticmethod
    def by_name(name):
        return Warehouse.objects.get(name=name)

    class Meta:
        ordering = [ 'name' ]

class Delta(models.Model):
    """Deltas are used to track all quantity adjustments to a part at a
    given warehouse.  The adjustment may be relative to existing
    quantity or absolute.
    """
    part = models.ForeignKey(Part, verbose_name='Part #',
                             help_text='The part whose quantity gets '
                             'adjusted.')
    is_absolute = models.BooleanField(verbose_name='Absolute Adjustment',
                                      default=False,
                                      help_text='If set, the adjustment is '
                                      'absolute otherwise it is relative.')
    adj = models.IntegerField(verbose_name='Adjustment Count',
                              help_text='The amount by which the part '
                               'quantity should be adjusted by.')
    txtn = models.ForeignKey('Transaction', verbose_name='Transaction #')
    warehouse = models.ForeignKey(Warehouse)

    def line_item(self):
        """For order or shipments items, return associated line-item."""
        line_items = Line_Item.objects.filter(txtn_id=self.txtn.id) \
                                      .filter(part_id=self.part_id)
        if len(line_items) != 1:
            return None
        return line_items[0]

    def line_cost(self):
        """For order or shipments items, return line-cost for this delta."""
        line_item = self.line_item()
        if line_item is None:
            return None
        return Decimal('%.2f' % (self.adj *
                                 float(line_item.line_cost) / line_item.qty))

    def __str__(self):
        return 'Delta %s' % self.id

class Assembly_Item(models.Model):
    assy = models.ForeignKey(Part, related_name='assembly_item_part',
                             verbose_name='Assembly Part #',
                             help_text='The part number of the assembly this '
                             'item belongs to.')
    comp = models.ForeignKey(Part, related_name='assembly_item_comp',
                             verbose_name='Component Part #',
                             help_text='The part number of the component of '
                             'this item.')
    qty = models.PositiveIntegerField(help_text='The quantity of this '
                                      'component required by the assembly.')
    refdes = models.TextField(blank=True, default='',
                              verbose_name='Refdes List',
                              help_text='A list of comma-separated reference '
                              'designators e.g., "R101,R304". '
                              'The length of this list should match Quantity.')

    def qty_with_overage(self, factor):
        q = math.ceil(abs(factor) * self.qty * (1 + self.comp.overage / 100))
        return int(math.copysign(q, factor))

    def __str__(self):
        return 'Assembly_Item %s.%s' % (format_part_number(self.assy_id),
                                        format_part_number(self.comp_id))

    def clean(self):
        refdes_list = re.findall(r'([^, ]+)', self.refdes)
        self.refdes = ','.join(refdes_list)

    class Meta:
        unique_together = ('assy', 'comp')
        index_together = [
            [ 'assy', 'comp' ]
        ]

class Line_Item(models.Model):
    txtn = models.ForeignKey('Transaction', verbose_name='Transaction #',
                              db_index=True)
    part = models.ForeignKey(Part, verbose_name='Part #')
    qty = models.PositiveIntegerField(help_text='The quantity of this part.')
    line_cost = models.DecimalField(max_digits = 9, decimal_places = 2,
                                    verbose_name='Amount',
                                    help_text='The total cost of this '
                                    'line-item.  Part-cost is line-item cost '
                                    'divided by quantity.',
                                    validators=[MinValueValidator(
                                        Decimal('0.00'))])
    index = models.PositiveIntegerField(help_text='Sequential and unchanging '
                                        'index of this line-item.')
    class Meta:
        ordering = [ 'index' ]
        unique_together = (
            ('txtn', 'part'),
            ('txtn', 'index')
        )
        index_together = [
            [ 'txtn', 'part' ]
        ]

    def vendor_part(self):
        result = Vendor_Part.objects \
                            .filter(part_id=self.part_id) \
                            .filter(vendor_id=self.txtn.order.vendor.id)
        if len(result) != 1:
            return None
        return result[0]

    def part_cost(self):
        return part_cost(self.line_cost, self.qty)

    def order_line_item(self):
        """For shipment line-items, return associated order line-item."""
        line_items = Line_Item.objects \
                              .filter(txtn_id=self.txtn.shipment.ordr.id) \
                              .filter(part_id=self.part_id)
        if len(line_items) != 1:
            return None
        return line_items[0]

    def qty_remaining_to_ship(self):
        if not hasattr(self.txtn, 'order'):
            return None

        shipment_ids = Shipment.objects.filter(ordr_id=self.txtn_id) \
                                       .values_list('transaction_ptr_id',
                                                    flat=True)
        already_shipped = sum(Line_Item.objects \
                              .filter(txtn_id__in=shipment_ids) \
                              .filter(part_id=self.part_id) \
                              .values_list('qty', flat=True))
        return self.qty - already_shipped

class Transaction(models.Model):
    """Orders, shipments, and inventories are all transactions."""
    ts = models.DateTimeField(verbose_name='Creation Time',
                              help_text='Date and time when this transaction '
                              'was created.')
    warehouse = models.ForeignKey(Warehouse,
                                  verbose_name='Warehouse',
                                  help_text='The (destination) warehouse '
                                  'this transaction applies to.')
    notes = models.TextField(blank=True,
                             verbose_name='Notes',
                             help_text='Comments and notes for this '
                             'transaction')

    def __str__(self):
        return 'Transaction %u' % (self.id)

    def first_available_index(self):
        max_index = 0
        for item in Line_Item.objects.filter(txtn_id=self.id):
            if item.index > max_index:
                max_index = item.index
        return max_index + 1

    def update_order_status(self):
        if hasattr(self, 'order'):
            order = self.order
        elif hasattr(self, 'shipment'):
            if self.ordr is None:
                return
            order = self.ordr
        else:
            return

        status = Order.STATUS_CLOSED
        for line_item in Line_Item.objects.filter(txtn_id=order.id):
            if line_item.qty_remaining_to_ship() > 0:
                status = Order.STATUS_OPEN
                break

        if status != order.status:
            order.status = status
            order.save()

    def update_deltas(self):
        if hasattr(self, 'order') or hasattr(self, 'inventory'):
            # orders never have deltas and inventories directly edit the deltas
            return

        # delete all existing deltas for this transaction:
        Delta.objects.filter(txtn_id=self.id).delete()

        if hasattr(self, 'shipment'):
            ship = self.shipment
            src_warehouse = ship.src_warehouse()

            for line_item in Line_Item.objects.filter(txtn_id=ship.id):
                d = Delta(part_id=line_item.part_id, is_absolute=False,
                          adj=line_item.qty, txtn_id=self.id,
                          warehouse_id=ship.warehouse_id)
                d.save()

                assy_items = d.part.assembly_items()
                if ship.ordr and assy_items.exists():
                    # assemblies ordered from an assembly-house are created
                    # from parts at the assembly-house:
                    for assy_item in assy_items:
                        qty = -assy_item.qty_with_overage(line_item.qty)
                        part = assy_item.comp.best_part ()
                        d = Delta(part_id=part.id, is_absolute=False,
                                  adj=qty, txtn_id=self.id,
                                  warehouse_id=src_warehouse.id)
                        d.save()
                elif ship.from_warehouse is not None:
                    d = Delta(part_id=line_item.part_id, is_absolute=False,
                              adj=-line_item.qty, txtn_id=self.id,
                              warehouse_id=src_warehouse.id)
                    d.save()

    def finalize(self):
        self.update_order_status()
        self.update_deltas()

    def description(self):
        if hasattr(self, 'order'):
            ttl = self.order.days_until_past_due()
            extra = ''
            if ttl and ttl <= 0:
                extra = ' (expected %d days ago)' % -ttl
            desc = '%s PO %s shipping to %s%s.' % \
                (self.order.vendor.html_link(), self.order.html_link(),
                 self.warehouse.html_link(), extra)
        elif hasattr(self, 'shipment'):
            ship = self.shipment
            if ship.ordr_id:
                desc = 'Shipment %s to %s for %s PO %s.' % \
                       (ship.html_link(), ship.warehouse.html_link(),
                        ship.ordr.vendor.html_link(), ship.ordr.html_link())
            else:
                desc = 'Shipment %s from %s to %s.' % \
                       (ship.html_link(), ship.from_warehouse.html_link(),
                        ship.warehouse.html_link())
        elif hasattr(self, 'inventory'):
            desc = 'Inventory %s at %s.' % (self.inventory.html_link(),
                                            self.warehouse.html_link())
        else:
            desc = 'Unknown transaction %d.' % self.id
        return desc

    def html_link(self):
        if hasattr(self, 'order'):
            return 'Order&nbsp;%s' % self.order.html_link()
        elif hasattr(self, 'shipment'):
            return 'Shipment&nbsp;%s' % self.shipment.html_link()
        elif hasattr(self, 'inventory'):
            return 'Inventory&nbsp;%s' % self.inventory.html_link()
        else:
            return 'Transaction&nbsp;%d' % (self.id)

class Order(Transaction):
    STATUS_OPEN = 0
    STATUS_CLOSED = 1
    STATUS_CHOICES = (
        (STATUS_OPEN,	'open'),
        (STATUS_CLOSED,	'closed'),
    )
    expected_arrival_date = models.DateField(verbose_name='Expected Arrival '
                                             'Date', help_text='Date when '
                                             'the order is expected to '
                                             'arrive.')
    status = models.IntegerField(choices=STATUS_CHOICES,
                                 default=STATUS_OPEN, blank=False,
                                 verbose_name='Order Status')
    vendor = models.ForeignKey(Vendor, help_text='The name of the '
                               'vendor (distributor) where the order '
                               'was placed.')

    def __str__(self):
        return 'Order %u' % (self.id)

    def total_cost(self):
        return sum(Line_Item.objects
                   .filter(txtn_id=self.id) \
                   .values_list('line_cost', flat=True))

    def is_open(self):
        return self.status == Order.STATUS_OPEN

    def days_until_past_due(self):
        if self.status != Order.STATUS_OPEN:
            return None
        return (self.expected_arrival_date - date.today()).days

    def assembly_line_items(self):
        """Returns the list of line-items that are assemblies."""
        assemblies = Assembly_Item.objects.values_list('assy_id', flat=True) \
                                          .distinct()
        return Line_Item.objects.filter(txtn_id=self.id) \
                                .filter(part_id__in=assemblies)

    def html_link(self):
        url = reverse('epic:order_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.id)

class Shipment(Transaction):
    tracking = models.CharField(max_length=127, default='', blank=True,
                                verbose_name='Tracking #s',
                                help_text='Comma-separated list of tracking '
                                'numbers.')
    # order is a reserved word in django...
    ordr = models.ForeignKey(Order, verbose_name='Order #',
                             null=True, blank=True,
                             help_text='For an order shipment, the '
                             'order that resulted in this shipment.')
    from_warehouse = models.ForeignKey(Warehouse, null=True, blank=True,
                                       help_text='For an inter-warehouse '
                                       'shipment, the warehouse '
                                       'the shipment originates from.')
    cost_freight = models.DecimalField(max_digits=9, decimal_places=2,
                                       verbose_name='Freight Cost',
                                       validators=[MinValueValidator(
                                           Decimal('0.00'))])
    cost_other = models.DecimalField(max_digits=9, decimal_places=2,
                                     verbose_name='Other Costs',
                                     help_text='Other costs assessed by the '
                                     'shipper, such as handling costs.',
                                     validators=[MinValueValidator(
                                         Decimal('0.00'))])
    cost_discount = models.DecimalField(max_digits=9, decimal_places=2,
                                        verbose_name='Discount Given',
                                        help_text='Discounts given by the '
                                        'shipper, such as early payment '
                                        'discount.',
                                        validators=[
                                            MinValueValidator(
                                                Decimal('0.00'))])

    costs = {}		# cache of cost-details

    @classmethod
    def flush_caches(cls, sender, **kwargs):
        """See signals.flush_caches()."""
        # cache of shipment costs, indexed by transaction id
        cls.costs = {}

    def __str__(self):
        return 'Shipment %u' % self.id

    def clean(self):
        tracking_list = re.findall(r'([^, ]+)', self.tracking)
        self.tracking = ','.join(tracking_list)

    def html_link(self):
        url = reverse('epic:ship_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.id)

    def html_tracking_links(self):
        trk_list = self.tracking.split(',')
        url_pattern = 'https://www.packagetrackr.com/track/%s'
        result = []
        for trk in trk_list:
            trk = trk.strip()
            url = url_pattern % trk
            result.append('<a target="trk_lookup" href="%s">%s</a>' \
                           % (url, trk))
        return html_list_or_none(result)

    def src_warehouse(self):
        if self.ordr is not None:
            try:
                return Warehouse.by_name(self.ordr.vendor)
            except:
                return None
        else:
            return self.from_warehouse

    def overhead_cost(self):
        """Return the cost of overheads of this shipment."""
        return self.cost_freight + self.cost_other - self.cost_discount

    def cost_details(self):
        if self.id not in Shipment.costs:
            parts_cost = 0
            parts_target_cost = 0
            part_detail = {}
            for line_item in Line_Item.objects.filter(txtn_id=self.id):
                if part_detail is not None:
                    part_detail[line_item.part_id] = {
                        'cost': line_item.line_cost,
                        'qty': line_item.qty
                    }
                if self.from_warehouse is None:
                    parts_cost += line_item.line_cost
                else:
                    # for inter-warehouse shipments, each line_item.line_cost
                    # is zero but we need an approximation of part cost so
                    # we can allocate the cost of the shipment properly
                    parts_target_cost += line_item.qty \
					 * line_item.part.target_price
            Shipment.costs[self.id] = {
                'overhead': self.overhead_cost(),
                'parts': parts_cost,
                'parts_target': parts_target_cost,
                'part_detail': part_detail
            }
        return Shipment.costs[self.id]

    def parts_cost(self):
        """Return the total cost of all parts in this shipment."""
        return self.cost_details()['parts']

    def total_cost(self):
        return self.overhead_cost() + self.parts_cost()

class Inventory(Transaction):
    def __str__(self):
        return 'Inventory %u' % self.id

    def html_link(self):
        url = reverse('epic:warehouse_inventory_detail',
                       kwargs={'warehouse': self.warehouse.id, 'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.id)



#
#   Copyright (c) 2014-2015 eGauge Systems LLC
# 	4730 Walnut St, Suite 110
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
import operator
import re

from copy import deepcopy

from datetime import datetime, timedelta

from django.conf import settings as cfg
from django.core.urlresolvers import reverse
from django.utils.html import escape
from django.utils import timezone

from epic.models import *

class error(Exception):
    pass

def get_assembly_name(part):
    """Get the canonical name for assembly PART.  We leave out
    MANUFACTURER for our own parts, but include the manufacturer name
    for uniqueness for others.

    """
    if part.mfg == cfg.EPIC_MANUFACTURER:
        return part.mfg_pn
    return part.mfg + ' ' + part.mfg_pn


def get_assemblies_using_part(part):
    """Get a list of assemblies that are using PART (or any of its
    substitutes).  Obsolete assemblies are ignored."""
    part_list = [ p.id for p in part.equivalent_parts() ]
    return Part.objects.filter(assembly_item_part__comp_id__in=part_list) \
                       .exclude(status=Part.STATUS_OBSOLETE) \
                       .distinct()

class Part_Count:
    def __init__(self, qty=0):
        self.qty = qty
        self.is_final = False

class Part_Inventory:
    def __init__(self):
        self.warehouse = {}

    def get_entry(self, warehouse_id, part_id):
        if warehouse_id in self.warehouse:
            entry = self.warehouse[warehouse_id]
        else:
            self.warehouse[warehouse_id] = entry = {}
        if part_id not in entry:
            entry[part_id] = Part_Count()
        return entry

    def apply_assembly_item(self, warehouse_id, item, factor):
        # Note: Even though this component might be an assembly itself,
        #  there is no need to recurse down because parts do not
        #  spontaneously transform into assemblies.  That is, even if
        #  we had the pre-requisite sub-components, there still has to
        #  be an assembly order which is where we'd do the proper
        #  sub-component qty adjustments.
        part_id = item.comp.best_part().id
        qty = item.qty_with_overage(factor)
        entry = self.get_entry(warehouse_id, part_id)
        entry[part_id].qty += qty

    def apply_order_line_item(self, item):
        part_id = item.part_id
        order = item.txtn.order
        qty = item.qty_remaining_to_ship()
        if qty <= 0:
            return
        entry = self.get_entry(order.warehouse_id, part_id)
        entry[part_id].qty += qty

        assy_items = item.part.assembly_items()
        if assy_items.exists() > 0:
            vendor_id = Warehouse.by_name(order.vendor).id
            for assy_item in item.part.assembly_items():
                self.apply_assembly_item(vendor_id, assy_item, -qty)

    def apply_delta(self, delta, reverse_chronological_order=False):
        """Adjust inventory according to DELTA.  If applying deltas in reverse
        chronological order (newest first), you must set
        REVERSE_CHRONOLOGICAL_ORDER to True and stop processing deltas
        after as the first delta with "is_absolute==True" has been
        processed.

        """
        entry = self.get_entry(delta.warehouse_id, delta.part_id)
        if entry[delta.part_id].is_final:
            return

        if delta.is_absolute:
            if reverse_chronological_order:
                entry[delta.part_id].qty += delta.adj
                entry[delta.part_id].is_final = True
            else:
                entry[delta.part_id].qty  = delta.adj
        elif not entry[delta.part_id].is_final:
                entry[delta.part_id].qty += delta.adj

    def get_qty(self, warehouse, part):
        if warehouse not in self.warehouse:
            return 0
        if part not in self.warehouse[warehouse]:
            return 0
        return self.warehouse[warehouse][part].qty

    def adj_qty(self, warehouse, part, adj):
        entry = self.get_entry(warehouse, part)
        entry[part].qty += adj

def get_stock(warehouses=None, as_of_date=None, exclude_txtn_id=None):
        """Get the current stock as a Part_Inventory instance for the
        specified list of warehouses.  If WAREHOUSES is None, the
        stock for all warehouses is returned.

        If AS_of_date is not None, only deltas with timestamps up
        to and including that time are considered.

        If EXCLUDE_TXTN_ID is not None, any deltas for that
        transaction id are ignored.
        """
        inv = Part_Inventory()

        qs = Delta.objects.order_by('-txtn__ts')
        if warehouses is not None:
            qs = qs.filter(warehouse__in=warehouses)
        if exclude_txtn_id is not None:
            qs = qs.exclude(txtn_id=exclude_txtn_id)
        if as_of_date is not None:
            as_of_ts = timezone.make_aware(
                datetime.combine(as_of_date, datetime.min.time()),
                timezone.utc)
            qs = qs.filter(txtn__ts__lte=as_of_ts)

        for d in qs:
            inv.apply_delta(d, reverse_chronological_order=True)

        if False:
            for w in inv.warehouse:
                s = ''
                for p in inv.warehouse[w]:
                    if inv.warehouse[w][p].qty != 0:
                        s += (' %d*%s' % (inv.warehouse[w][p].qty,
                                          format_part_number(p)))
                print('Warehouse %s stock:%s' % (w, s))
        return inv

def get_stock_summary(warehouse=None, as_of_date=None):
    """Get a summary of the stock at WAREHOUSE (or all of them if
    WAREHOUSE is None).  Returns a list and the total cost of the
    parts in stock.  The list contains dictionaries that break down
    the parts inventory: part_list is the list of equivalent parts and
    their sub-quantities, total_qty is the total_qty (across
    equivalent parts), 'price' is the average price of the best part
    and 'amount' is simply 'price' * 'total_qty'.

    """
    inv = get_stock([ warehouse ] if warehouse else None,
                    as_of_date=as_of_date)

    if warehouse is None:
        parts = {}
        for w in inv.warehouse:
            for p in inv.warehouse[w]:
                if p not in parts:
                    parts[p] = Part_Count(0)
                parts[p].qty += inv.warehouse[w][p].qty
    elif warehouse.id in inv.warehouse:
        parts = inv.warehouse[warehouse.id]
    else:
        parts = {}

    best_parts = {}
    for p in parts:
        part = Part.objects.get(pk=p)
        best = part.best_part()
        if best.id in best_parts:
            continue
        best_parts[best.id] = best

    stock = []
    stock_total = 0
    for best_part_id, best_part in best_parts.items():
        qty = parts[best_part_id].qty if best_part_id in parts else 0
        total_qty = qty

        # always list best part first:
        part_list = [ ( best_part,  qty ) ]
        for p in best_part.equivalent_parts():
            if p == best_part or p.id not in parts:
                continue
            qty = parts[p.id].qty
            part_list.append(( p, qty ) )
            total_qty += qty

	# do not suppress negative counts, as those are useful to be seen
        if total_qty == 0:
            continue

        price = best_part.avg_cost()
        amount = total_qty * price
        stock_total += amount
        stock.append({ 'part_list': part_list, 'total_qty': total_qty,
                       'price': price, 'amount': amount})
    return stock, stock_total

def get_inventory_summary(warehouse_id, inventory):
    inv = get_stock(warehouses=[ warehouse_id ], as_of_date=inventory.ts,
                    exclude_txtn_id=inventory.id)

    total_value_change = 0
    has_relative_deltas = False
    inv_items = {}
    for d in Delta.objects.filter(txtn_id=inventory.id):
        best_part = d.part.best_part()

        if best_part.id in inv_items:
            item = inv_items[best_part.id]
        else:
            item = {
                # relative inventory items are no longer supported:
                'is_relative':  False,
                'part':		best_part,
                'value_diff':	0,
                'qty_old_list':	[],
                'qty_new_list': [],
                'qty_dif_list': []
            }
            inv_items[best_part.id] = item
        old_qty = inv.get_qty(warehouse_id, d.part_id)
        if d.is_absolute:
            new_qty = d.adj
        else:
            new_qty = old_qty + d.adj
            item['is_relative'] = True
            has_relative_deltas = True
        dif_qty = new_qty - old_qty
        dif_val = dif_qty * d.part.avg_cost()
        item['qty_old_list'].append( (old_qty, d.part) )
        item['qty_new_list'].append( (new_qty, d.part) )
        item['qty_dif_list'].append( (dif_qty, d.part) )
        item['value_diff'] += dif_val
        total_value_change += dif_val

    inventory_list = []
    for part, item in inv_items.items():
        inventory_list.append (item)
    inventory_list.sort(key=lambda x: x['part'].id)
    inventory_list.sort(key=operator.itemgetter('value_diff'))
    return inventory_list, total_value_change, has_relative_deltas

def get_open_order_line_items(last_order=None):
    """Return query-set of line-items on open orders.  If LAST_ORDER is
    not None, do not consider any orders that are expected to arrive
    later than that LAST_ORDER.

    """
    qs = Line_Item.objects.filter(txtn__order__status=Order.STATUS_OPEN)
    if last_order is not None:
        qs = qs.exclude(txtn__order__expected_arrival_date__gt=\
                        last_order.expected_arrival_date)
    return qs

def get_open_order_summary(last_order=None):
    """Return a list of open orders, sorted by expected arrival date.  If
       LAST_ORDER is not None, orders after the expected arrival date
       of LAST_ORDER are not considered.

       For each open order, return a dictionary with the order ('order')
       and the list of line-items in that order ('items').
    """
    qs = get_open_order_line_items(last_order)
    txtns = []
    prev_txtn = None
    for item in qs.order_by('txtn_id'):
        if prev_txtn != item.txtn_id:
            prev_txtn = item.txtn_id
            entry = {
                'order': item.txtn.order,
                'items': []
            }
            txtns.append(entry)
        entry['items'].append(item)
    txtns.sort(key=lambda x: x['order'].expected_arrival_date)
    return txtns

class Part_Event:
    KIND_CURRENT_INVENTORY = 'current inventory'
    KIND_EXPECTED_SHIPMENT = 'expected shipment'
    KIND_ABSOLUTE_COUNT	   = 'count'

    def __init__(self, event_obj, ts, overdue=False):
        self.event_obj = event_obj
        self.ts = ts
        self.warehouse = None
        self.txtn = None
        self.kind = None
        self.part = None
        self.desc = None
        self.inv  = None
        self.part = None
        self.part_qty = None
        self.overdue = overdue

    def __str__(self):
        return 'obj %s ts %s whouse %s txtn %s kind %s part %s desc %s' % \
            (self.event_obj, self.ts, self.warehouse, self.txtn, self.kind,
             self.part, self.desc)

def get_part_history(part):
    """Get the purchase history for PART.  Returns a list of Part_Event
       instances ordered by descending timestamp.  Each Part_Event has
       a timestamp (ts) when the event occurred, the number of the
       transaction that resulted in the event (txtn), a string
       describing the kind of event it was (kind), a verbose
       descrption of the event (desc), and the inventory-level *after*
       the event occurred (inv).

    """
    part_list = [ p.id for p in part.equivalent_parts() ]
    open_order_list = [ o.txtn_id for o in get_open_order_line_items() ]
    event = []

    # List of assembly IDs that use any parts in PART_LIST:
    assy_ids = Assembly_Item.objects.filter(comp_id__in=part_list) \
                                    .values_list('assy_id', flat=True)

    part_or_assy_list = part_list[:]
    part_or_assy_list.extend(assy_ids)

    inv = Part_Inventory()

    # get open orders for the part:
    for item in Line_Item.objects.filter(part_id__in=part_or_assy_list) \
                                 .filter(txtn_id__in=open_order_list):
        order = item.txtn.order

        # sort overdue orders as if they arrive tomorrow:
        dt = order.expected_arrival_date
        ts = timezone.make_aware(datetime.combine(dt, datetime.min.time()),
                                  timezone.utc)
        limit = timezone.now() + timedelta(days=1)
        overdue = False
        if ts < limit:
            overdue = True
            ts = limit

        if item.qty_remaining_to_ship() > 0:
            if item.part_id in assy_ids:
                qs = item.part.assembly_items().filter(comp_id__in=part_list)
                if not qs.exists():
                    continue
                comp = None
                qty = 0
                for assy_item in qs:
                    comp = assy_item.comp
                    qty += assy_item.qty_with_overage(item.qty)
                pe = Part_Event(item, ts, overdue=overdue)
                pe.part = comp.best_part()
                pe.part_qty = qty
            else:
                pe = Part_Event(item, ts, overdue=overdue)
            event.append(pe)

    # get deltas for the part:
    for delta in Delta.objects.filter(part_id__in=part_list):
        event.append(Part_Event(delta, delta.txtn.ts))

    todays_inventory = Part_Event(None, timezone.now())
    todays_inventory.kind = Part_Event.KIND_CURRENT_INVENTORY
    event.append(todays_inventory)

    # sort from oldest to newest:
    event.sort(key=lambda pe: pe.ts)

    for pe in event:
        if pe.event_obj is None:
            pass
        elif isinstance(pe.event_obj, Line_Item) \
             and hasattr(pe.event_obj.txtn, 'order'):
            inv.apply_order_line_item(pe.event_obj)
            item = pe.event_obj
            txtn = order = item.txtn.order
            order_url = reverse('epic:order_detail',
                                 kwargs={'pk': item.txtn_id})

            if pe.part_qty:
                part_html = html_part_link(pe.part)
                pe.desc = 'Expect use of %d&times;%s by %s ' \
                          '<a href="%s">order</a> placed on %s.' % \
                          (pe.part_qty, part_html, order.vendor, order_url,
                           order.ts.strftime('%x'))
                if pe.overdue:
                    pe.desc += '<br>Shipment is overdue!  ' \
                               'Originally expected on %s.' \
                               % (order.expected_arrival_date.strftime('%x'))
                pe.warehouse = Warehouse.by_name(order.vendor)
            else:
                part_html = html_part_link(item.part_id)
                pe.desc = 'Expect delivery of %d&times;%s from %s ' \
                          '<a href="%s">order</a> placed on %s, ' \
                          'shipping to %s.' % \
                          (item.qty, part_html, order.vendor, order_url,
                           order.ts.strftime('%x'),
                           order.warehouse.name)
                pe.warehouse = order.warehouse_id
            if txtn.notes:
                pe.desc += '<br><em>%s</em>' % txtn.notes
            pe.txtn = txtn.id
            pe.kind = Part_Event.KIND_EXPECTED_SHIPMENT
            pe.part = item.part_id
        elif isinstance(pe.event_obj, Delta):
            delta = pe.event_obj
            inv.apply_delta(delta)
            part_html = html_part_link(delta.part_id)
            txtn = delta.txtn
            omit_dst_warehouse = False
            if hasattr(txtn, 'order'):
                raise error('Unexpected order %s' % txtn.id)
            elif hasattr(txtn, 'shipment'):
                kind = 'shipment'
                txtn_url = reverse('epic:ship_detail', kwargs={'pk': txtn.id})
                ship = txtn.shipment
                if delta.adj < 0:
                    if ship.ordr is not None:
                        desc = '<a href="%s">Used</a> %u&times;%s at' % \
                               (txtn_url, -delta.adj, part_html)
                    else:
                        desc = '<a href="%s">Sent</a> %u&times;%s from %s' % \
                               (txtn_url, -delta.adj, part_html,
                                ship.from_warehouse)
                        omit_dst_warehouse = True
                else:
                    if ship.ordr is not None:
                        url = reverse('epic:order_detail',
                                      kwargs={'pk': ship.ordr_id})
                        frm = '%s <a href="%s">order</a>' \
                              % (ship.ordr.vendor, url)
                        desc = '<a href="%s">Shipped</a> %u&times;%s ' \
                               'from %s to' % \
                               (txtn_url, delta.adj, part_html, frm)
                    else:
                        desc = '<a href="%s">Received</a> %u&times;%s at' % \
                               (txtn_url, delta.adj, part_html)
            elif hasattr(txtn, 'inventory'):
                kind = 'inventory'
                txtn_url = reverse('epic:warehouse_inventory_detail',
                                   kwargs={'warehouse': txtn.warehouse.id,
                                           'pk': txtn.id })
                if delta.is_absolute:
                    desc = '<a href="%s">Counted</a> %u&times;%s at' % \
                           (txtn_url, delta.adj, part_html)
                    kind = Part_Event.KIND_ABSOLUTE_COUNT
                else:
                    desc = '<a href="%s">Adjusted</a> by %d&times;%s at' % \
                           (txtn_url, delta.adj, part_html)
            else:
                raise error('Unexpected transaction %s' % txtn.id)
            if not omit_dst_warehouse:
                desc += ' %s' % delta.warehouse.name
            desc += '.'
            if txtn.notes:
                desc = desc + '<br><em>%s</em>' % \
                       txtn.notes
            pe.warehouse = delta.warehouse_id
            pe.txtn = txtn.id
            pe.kind = kind
            pe.part = delta.part_id
            pe.desc = desc
        else:
            raise error('Unexpected event object %s' % str(pe.event_obj))
        pe.inv = deepcopy(inv)

    todays_inventory.desc = 'Current warehouse stock as of today.'

    # Now reverse so that we have newest to oldest:
    event.reverse()
    return event

def part_history_summary(part, events, full_history=False):
    def warehouse_done(e, w):
        if w not in e.inv.warehouse:
            return True
        if w not in warehouse_absolute:
            return False
        return all(part in warehouse_absolute[w] for part in part_list)

    def warehouse_part_done(w, p):
        return w in warehouse_absolute and p in warehouse_absolute[w]

    def update_warehouse_absolute(e):
        if full_history:
            return
        if isinstance(e.event_obj, Delta) and e.event_obj.is_absolute:
            if e.warehouse not in warehouse_absolute:
                warehouse_absolute[e.warehouse] = {}
            warehouse_absolute[e.warehouse][e.part] = True

    part_list = [ p.id for p in part.equivalent_parts() ]

    # Only show warehouses that have non-zero part counts or expected
    # shipments in the period that we display:
    warehouse_absolute = {}
    warehouses_to_show = {}
    for e in events:
        if warehouse_done(e, e.warehouse):
            continue
        if e.kind == Part_Event.KIND_EXPECTED_SHIPMENT:
            warehouses_to_show[e.warehouse] = True
        else:
            for w in e.inv.warehouse:
                if warehouse_done(e, w):
                    continue
                part_qtys = e.inv.warehouse[w]
                for p in part_list:
                    if p not in part_qtys:
                        continue
                    if part_qtys[p].qty != 0:
                        if not warehouse_part_done(w, p):
                            warehouses_to_show[w] = True
        update_warehouse_absolute(e)
    for w in warehouses_to_show:
        warehouses_to_show[w] = Warehouse.objects.get(pk=w)

    now = timezone.now()
    first_present = True
    warehouse_absolute = {}
    event_summary = []
    for e in events:
        if (e.kind != Part_Event.KIND_CURRENT_INVENTORY
            and e.kind != Part_Event.KIND_EXPECTED_SHIPMENT
            and (e.warehouse not in warehouses_to_show
                 or warehouse_part_done(e.warehouse, e.part))):
            continue

        summary = {}
        cls = ''
        if e.ts > now:
            cls = 'danger' if e.overdue else 'warning'
        elif first_present and e.ts <= now:
            cls = 'success'
            first_present = False
        summary['class'] = cls
        summary['ts'] = e.ts
        summary['desc'] = e.desc
        summary['warehouses'] = []
        for w in warehouses_to_show:
            items = []
            if w in e.inv.warehouse:
                part_qtys = e.inv.warehouse[w]
                for p in part_list:
                    if p not in part_qtys or warehouse_part_done(w, p):
                        continue
                    part = Part.objects.get(pk=p)
                    items.append({ "qty": part_qtys[p].qty, "part": part })
            summary['warehouses'].append(items)
        event_summary.append(summary)
        update_warehouse_absolute(e)

    return {
        "warehouses": warehouses_to_show.values(),
        "events": event_summary
    }

def page_sibling(model, pk, backward):
    op_name = 'pk__' + ('lt' if backward else 'gt')
    qs = model.objects.order_by('-pk' if backward else 'pk') \
                      .filter(**{ op_name: pk })[:1] \
                      .values_list('pk', flat=True)
    return qs[0] if qs.exists() else None

def html_page_sibling(url_name, model, pk, backward):
    sibling_pk = page_sibling(model, pk, backward)
    html = '<a class="btn btn-default" '
    if sibling_pk is None:
        html += 'disabled="disabled"'
    else:
        url = reverse(url_name, kwargs={'pk': sibling_pk})
        html += 'href="%s"' % (url)
    html += '>'
    if backward:
        btn_dir = 'Prev.'
        icon_name = 'left'
    else:
        btn_dir = 'Next'
        icon_name = 'right'
    html += '%s %s <i class="glyphicon glyphicon-chevron-%s">' \
            % (btn_dir, model.__name__, icon_name)
    html += '</i></a>'
    return html

def html_page_nav(url_name, model, pk):
    """Create the 'Previous' and 'Next' navigation buttons for a paginator."""
    html  = html_page_sibling(url_name, model, pk, backward=True)
    html += ' '
    html += html_page_sibling(url_name, model, pk, backward=False)
    return html

def html_list_pager(page_obj, num_neighbors=5, page_name=lambda pg: pg,
                    queries=None, key='page'):
    """Given a Django Page object, return a bootstrap pageinator which
    provides direct access for a couple (NUM_NEIGHBORS) of neighboring
    pages.  If not all pages are directly accessible, ellipses are
    used to provide access on a neighbor-group by neighbor-group
    basis.  PAGE_NAME can be used to translate the page number to a custom
    page name.

    """
    def url(page_number):
        if page_number < 1 or page_number > num_pages:
            return ""
        q = queries[:]
        q.append('%s=%d' % (key, page_number))
        return '?' + "&".join(q)

    if queries is None:
        queries=[]

    num_pages = page_obj.paginator.num_pages

    if num_pages <= 1:
        return ""
    html  = '<nav><ul class="pagination">'
    # previous-page button:
    prev_symbol = '<span aria-hidden="true">&laquo;</span>'
    if page_obj.number <= 1:
        html += '<li class="disabled">%s</li>' % (prev_symbol)
    else:
        html += '<li><a href="%s">%s</a></li>' % (url(page_obj.number - 1),
                                                  prev_symbol)
    first_page = page_obj.number - num_neighbors
    if first_page < 1:
        first_page = 1
    if first_page > 1:
        right = first_page - 1
        left  = right - (num_neighbors - 1)
        if left < 1:
            left = 1
        ellipsis_page = (right + left) / 2
        html += '<li><a href="%s">&hellip;</a></li>' % (url(ellipsis_page))
        first_page += 1 # ellipsis occupies one of the neighbor blocks...
    for pg in range(first_page, page_obj.number):
        html += '<li><a href="%s">%s</a></li>' % (url(pg), page_name(pg))
    html += '<li class="active"><a>%s</a></li>' % page_name(page_obj.number)

    last_page = page_obj.number + num_neighbors
    if last_page > num_pages:
        last_page = num_pages
    if last_page < num_pages:
        left  = last_page + 1
        right = left + (num_neighbors - 1)
        if right > num_pages:
            right = num_pages
        ellipsis_page = (right + left) / 2
        last_page -= 1 # ellipsis occupies one of the neighbor blocks...

    for pg in range(page_obj.number + 1, last_page + 1):
        html += '<li><a href="%s">%s</a></li>' % (url(pg), page_name(pg))
    if last_page < num_pages:
        html += '<li><a href="%s">&hellip;</a></li>' % (url(ellipsis_page))

    # next-page button:
    next_symbol = '<span aria-hidden="true">&raquo;</span>'
    if page_obj.number >= num_pages:
        html += '<li class="disabled">%s</li>' % (next_symbol)
    else:
        html += '<li><a href="%s">%s</a></li>' % (url(page_obj.number + 1),
                                                  next_symbol)
    html += '</ul></nav>'
    return html

def get_field(cls, field_name):
    for f in cls._meta.fields:
        if f.name == field_name:
            return f
    return None

def get_model_fields(obj, fields,
                     map = lambda name, val, vname: (vname, val)):
    """Get the model fields and values for OBJ.  Any model fields not
    listed in FIELDS are skiiped.  The values are HTML-escaped so they
    are safe to be output directly.
    """
    field_list = []
    for field_name in fields:
        f = get_field(obj.__class__, field_name)
        if f is None:
            continue
        if f.name + '_ptr_id' in obj.__dict__:
            v = obj.__dict__[f.name + '_ptr_id']
        elif f.name + '_id' in obj.__dict__:
            v = obj.__dict__[f.name + '_id']
        else:
            v = obj.__dict__[f.name]
        get_name = 'get_' + f.name + '_display'
        if hasattr(obj, get_name):
            v = getattr(obj, get_name)()
        vname = f.verbose_name
        if not vname[0].isupper():
            vname = vname.capitalize()
        name, v = map(f.name, escape(v), vname)
        field_list.append([name, v])
    return field_list

def get_initial_from_post(request, item_prefix, txtn_model, item_model):
    """Parses the POST data from a form that provides initial data for a
    transaction object (e.g., Order or Shipment) and the associated
    items.  ITEM_PREFIX is the prefix used for the items, TXTN_MODEL is
    the Django model for the transaction and ITEM_MODEL tha of the items.
    These are used to convert the form data to the Python data needed for
    initialization.  Returns two lists: the initial data of the transaction
    and that of the items.

    """
    prefix_pattern = re.compile(r'^%s-(\d+)-(.*)' % item_prefix)
    items = {}
    txtn_initial = {}
    for name, value in request.POST.items():
        m = prefix_pattern.match(name)
        if m:
            idx = m.group(1)
            field_name = m.group(2)
            if idx not in items:
                items[idx] = {}
            for f in item_model._meta.fields:
                if f.name == field_name:
                    items[idx][field_name] = f.to_python(value)
        else:
            for f in txtn_model._meta.fields:
                if f.name == name:
                    txtn_initial[name] = f.to_python(value)
    items_initial = []
    for key, val in items.items():
        items_initial.append(val)
    return txtn_initial, items_initial

def breadcrumb(path):
    if path[len(path) - 1] == '/':
        path = path[:-1]
    comps = path.split('/')
    content = ''
    for i in range(len(comps)):
        if i == 0:
            continue
        else:
            path = '/'.join(comps[0:i+1])
            name = comps[i]
            if name == 'epic':
                name = 'EPIC'
            elif name == 'bom':
                name = 'BOM'
            else:
                name = name.capitalize()
        if i == len(comps) - 1:
            content += '<li class="active">%s</li>' % name
        else:
            content += '<li><a href="%s">%s</a></li>' % (path, name)
    return ('<ol class="breadcrumb">%s</ol>' % content)



#
#   Copyright (c) 2014-2015 eGauge Systems LLC
# 	4730 Walnut St, Suite 110
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
import autocomplete_light
import crispy_forms.helper
import json
import math
import operator
import os
import re

from autocomplete_light import TextWidget as AutocompleteTextWidget
from bootstrap3_datetime.widgets import DateTimePicker
from collections import OrderedDict
from datetime import date, datetime, timedelta

from django import forms
from django.db import transaction
from django.conf import settings as cfg
from django.contrib.auth.decorators import permission_required
from django.contrib.humanize.templatetags.humanize import intcomma
from django.contrib.staticfiles import finders
from django.core.exceptions import ObjectDoesNotExist, ValidationError, \
    PermissionDenied
from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage
from django.core.urlresolvers import reverse
from django.forms import ModelForm
from django.forms.forms import NON_FIELD_ERRORS
from django.forms.formsets import DELETION_FIELD_NAME
from django.forms.fields import CharField
from django.forms.models import inlineformset_factory, BaseInlineFormSet
from django.forms.util import ErrorList
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import render, get_object_or_404
from django.template import defaultfilters
from django.utils import timezone
from django.views import generic

from epic import perms
from epic.apps import EPIC_App_Config
from epic.footprints import Footprints
from epic.lib import *
from epic.models import *

def crispy_form_helper():
    helper = crispy_forms.helper.FormHelper()
    helper.form_class = 'form-horizontal'
    helper.label_class = 'col-md-3'
    helper.field_class = 'col-md-9'
    helper.form_tag = False
    return helper

def crispy_inline_form_helper():
    helper = crispy_forms.helper.FormHelper()
    # See https://github.com/maraujop/django-crispy-forms/issues/376
    # as to why we can't used bootstrap3/table_inline_formset.html
    # directly.  It results in a mangled 'Delete' checkbox column.
    helper.template = 'epic/inline_formset.html'
    helper.form_tag = False
    return helper

def assert_permission(request, perm):
    if not request.user.has_perm(perm):
        raise PermissionDenied

class Part_Autocomplete(autocomplete_light.AutocompleteModelBase):
    search_fields=[ '^mfg', 'mfg_pn' ]
    attrs = {
        'placeholder': 'Part #?',
        'data-autocomplete-minimum-characters': 1
    }
    widget_attrs={
        'data-widget-maximum-values': 8,
    }

    def choices_for_request(self):
        assert_permission(self.request, perms.VIEW)
        choices = self.choices.all()

        pid = self.request.GET.get('q', '')
        if re.match(r'\d+$', pid, re.I):
            # if search if for a number, return that part alone:
            qs = self.choices.filter(id=pid)
            if len(qs) == 1:
                return qs
        return super(Part_Autocomplete, self).choices_for_request()

    def choice_label(self, choice):
        return '%s: %s %s' % (choice, choice.mfg, choice.mfg_pn)

    def choice_value(self, choice):
        return '%d' % choice.id
autocomplete_light.register(Part, Part_Autocomplete)

class Order_Autocomplete(autocomplete_light.AutocompleteModelBase):
    order_by = '-id'
    search_fields=[ '^vendor__name', 'warehouse__name' ]
    attrs = {
        'placeholder': 'Order number?',
        'data-autocomplete-minimum-characters': 1
    }
    widget_attrs={
        'data-widget-maximum-values': 8,
    }

    # It would be nice to limit the choices to just open orders, but
    # that doesn't work when we open an existing shipment which is
    # associated with a closed order.  In that case, we lose the order
    # because that closed order won't be a valid choice...
    #choices = Order.objects.filter (status=Order.STATUS_OPEN)
    choices = Order.objects.all()

    def choices_for_request(self):
        assert_permission(self.request, perms.VIEW)
        tid = self.request.GET.get('q', '')
        qs = Order.objects.none()
        if re.match(r'\d+$', tid):
            # tid is numeric: search open orders for transaction
            # numbers containing that numeric string:
            for c in self.choices.all():
                if re.match('\d*%s\d*' % tid, '%d' % c.id):
                    qs |= Order.objects.filter(id=c.id)
            if qs.exists():
                return qs
        return super(Order_Autocomplete, self).choices_for_request()

    def choice_label(self, choice):
        return '%s: %s $%s' % (choice.id, choice.vendor.name,
                               intcomma(choice.total_cost()))
autocomplete_light.register(Order, Order_Autocomplete)

class Manufacturer_Autocomplete(autocomplete_light.AutocompleteListBase):
    attrs = {
        'placeholder': 'Manufacturer?',
        'data-autocomplete-minimum-characters': 1
    }
    widget_attrs={
        'data-widget-maximum-values': 8,
    }

    choices = [ ]

    def choices_for_request(self):
        assert_permission(self.request, perms.VIEW)
        choices = Part.objects.values_list('mfg', flat=True).distinct()
        Manufacturer_Autocomplete.choices = choices
        return super(Manufacturer_Autocomplete, self).choices_for_request()
autocomplete_light.register(Manufacturer_Autocomplete)

class Footprint_Autocomplete(autocomplete_light.AutocompleteListBase):
    attrs = {
        'placeholder': 'Footprint?',
        'data-autocomplete-minimum-characters': 1
    }
    widget_attrs={
        'data-widget-maximum-values': 8,
    }

    choices = Footprints.get()

    def choices_for_request(self):
        assert_permission(self.request, perms.VIEW)
        # if len(Footprint_Autocomplete.choices) == 0:
        #     Footprint_Autocomplete.choices = Footprints.get()
        return super(Footprint_Autocomplete, self).choices_for_request()
autocomplete_light.register(Footprint_Autocomplete)

@permission_required(perms.VIEW)
def part_info(request):
    """This lets Javascript query for part info.  The query-format is:

    /epic/part/info/?pid=PART_ID&vid=VENDOR_ID

    where PID is the part id and VID is the id of the vendor.  If only
    PID is specified, the part-id (pid), manufacturer (mfg),
    manufacturer's part-number (mfg_pn), and the target-price (price)
    are returned.  The values are returned as a dictionary, indexed by
    the part-number.

    If PID and VID are specified and the part is available from the
    specified vendor, additionally the vendor's part-number (vendor_pn)
    is returned.  Also, instead of the part's target-price, the
    vendor's price is returned (price).

    PID may also be a comma-separated list of part-ids, in which case
    info for each of those parts is returned.

    If only VID is specified, all that vendor's parts are returned.

    """
    pid = request.GET.get('pid', '')
    if pid:
        try:
            pid_list = [ int(pid_str) for pid_str in pid.split(',')]
        except ValueError:
            pid = None
    vid = request.GET.get('vid', None)
    qs = Part.objects.none()
    if vid:
        try:
            qs = Vendor_Part.objects.filter(vendor_id=int(vid))
            if pid:
                qs = qs.filter(part_id__in=pid_list)
                if len(qs) == 0:
                    qs = Part.objects.filter(id__in=pid_list)
        except ValueError:
            vid = None
    elif pid:
        qs = Part.objects.filter(id__in=pid_list)

    info = {}
    for part in qs:
        i = {}
        if isinstance(part, Vendor_Part):
            vp = part
            part = vp.part
            i['vendor_pn'] = vp.vendor_pn
            i['price'] = '%s' % vp.price
        else:
            i['price'] = '%s' % part.target_price
        i['mfg'] = part.mfg
        i['mfg_pn'] = part.mfg_pn
        info[part.id] = i
    json_info = json.dumps(info)
    return HttpResponse(json_info, content_type='application/json')

class SubstituteWidget(AutocompleteTextWidget):

    def __init__(self, *args, **kwargs):
        if 'help_text' in kwargs:
            del kwargs['help_text']
        self.input_name = ''
        super(SubstituteWidget, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs=None):
        input = super(SubstituteWidget, self).render(name, None, attrs)
        part = self.part_instance
        sub_list = []
        part_attr = ''
        if part is not None:
            for p in part.equivalent_parts():
                if p.id != part.id:
                    sub_list.append('%d' % p.id)
            sub_list.sort(key=lambda x: int(x))
            if part.id is not None:
                part_attr = 'part="%d"' % part.id
        return('<span class="form-control epic_substitute_deck">'
               ' <span class="epic_substitute_list" style="display:none" %s>'
               '%s</span>'
               ' <span class="epic_substitute_add btn btn-default"'
               '    style="text-align:start">'
               '  %s'
               '  <input type="hidden" name="%s">'
               ' </span>'
               '</span>') \
            % (part_attr, ','.join(sub_list), input, self.input_name)

    def build_attrs(self, extra_attrs=None, **kwargs):
        attrs = super(SubstituteWidget, self).build_attrs(extra_attrs,
                                                          **kwargs)
        # remove the form-control class from the <input> since that
        # messes up formatting:
        if 'class' in attrs:
            cls = attrs['class']
            attrs['class'] = re.sub(r'(^| )form-control($| )', ' ', cls, re.I)
        if 'instance' in attrs:
            self.part_instance = attrs['instance']
            del attrs['instance']
        if 'name' in attrs:
            self.input_name = attrs['name']
            del attrs['name']
        self.input_attrs = attrs
        return attrs

class SubstituteField(CharField):

    def __init__(self, *args, **kwargs):
        if not 'widget' in kwargs:
            kwargs['widget'] = SubstituteWidget('Part_Autocomplete', **kwargs)
        super(SubstituteField, self).__init__(*args, **kwargs)

    def validate(self, value):
        list = value.split(',')
        instance = self.widget.attrs.get('instance')
        equiv = None if instance is None else instance.equivalent_parts()
        for part_str in list:
            if len(part_str) == 0:
                continue
            pn = int(part_str)
            params = {'part': format_part_number(pn) }
            try:
                p = Part.objects.get(pk=pn)
            except:
                raise ValidationError(
                    'Invalid part-number: %(part)s', params=params)
            if instance and p.id == instance.id:
                raise ValidationError(
                    'Part %(part)s cannot be its own substitute',
                    params=params)
            if p not in equiv and len(p.equivalent_parts()) != 1 \
               and len(equiv) != 1:
                raise ValidationError(
                    'Part %(part)s is already in a different '
                    'substitute set.  Remove it from that set '
                    'before adding it to this set.',
                    params=params)

class Part_Edit_Form(autocomplete_light.ModelForm):
    substitute = SubstituteField(help_text=get_field(Part, 'substitute') \
                                 .help_text)

    def __init__(self, *args, **kwargs):
        super(Part_Edit_Form, self).__init__(*args, **kwargs)
        self.helper = crispy_form_helper()
        self.fields['substitute'].widget.attrs['instance'] = self.instance

    def clean_footprint(self):
        return self.cleaned_data.get('footprint').strip()

    def clean_substitute(self):
        my_substitute = None
        old_set = set()
        new_set = set()
        if self.instance is not None and self.instance.id is not None:
            my_substitute = self.instance.substitute
            old_set = self.instance.equivalent_parts()
            new_set = set([ self.instance ])
        for sub in self.cleaned_data['substitute'].split(','):
            if len(sub) == 0:
                continue
            new_set.add(get_object_or_404(Part, pk=int(sub)))
        added = new_set - old_set
        deleted = old_set - new_set
        current = old_set
        changed = set()
        for d in deleted:
            current.remove(d)
            for prev in current:
                if prev.substitute == d:
                    prev.substitute = d.substitute
                    changed.add(prev)
            if d.substitute is not None:
                d.substitute = None
                changed.add(d)
        for a in added:
            if (self.instance is None or self.instance.id is None) and \
               my_substitute is None:
                my_substitute = a
            else:
                for prev in current:
                    if prev.substitute is None:
                        prev.substitute = a
                        changed.add(prev)
            current.add(a)

        for c in changed:
            if self.instance is not None and self.instance.id is not None \
               and c.id == self.instance.id:
                my_substitute = c.substitute
            else:
                c.save()
        return my_substitute

    class Meta:
        model = Part
        fields = [ 'mfg', 'mfg_pn', 'descr', 'val', 'footprint', 'mounting',
                   'target_price', 'overage', 'spq', 'lead_time',
                   'status', 'substitute' ]
        autocomplete_names = { 'footprint': 'Part_Autocomplete' }
        widgets = {
            'mfg': AutocompleteTextWidget('Manufacturer_Autocomplete'),
            'footprint': AutocompleteTextWidget('Footprint_Autocomplete')
        }

class Vendor_Part_Edit_Formset(BaseInlineFormSet):
    def __init__(self, *args, **kwargs):
        super(Vendor_Part_Edit_Formset, self).__init__(*args, **kwargs)
        self.helper = crispy_inline_form_helper()

    def add_fields(self, form, index):
        super(Vendor_Part_Edit_Formset, self).add_fields(form, index)
        form.fields['status'].help_text = ''

    def clean(self):
        super(Vendor_Part_Edit_Formset, self).clean()

        for form in self.forms:
            vendor = form.cleaned_data.get('vendor')
            if vendor and form.cleaned_data.get(DELETION_FIELD_NAME):
                qs = Order.objects.filter(vendor_id=vendor.id) \
                                  .filter(line_item__part_id=self.instance.id)
                if qs.exists():
                    errors = form._errors.setdefault('vendor_pn',
                                                     ErrorList())
                    errors.append('Vendor Part # cannot be deleted as it is '
                                  'referenced by one or more orders.')
                    raise ValidationError('Vendor Part # in use')

class Vendor_Part_Edit_Form(ModelForm):
    class Meta:
        model = Vendor_Part
        fields = [ 'vendor', 'vendor_pn', 'price', 'status' ]

    def __init__(self, *args, **kwargs):
        super(Vendor_Part_Edit_Form, self).__init__(*args, **kwargs)
        self.fields['vendor_pn'].widget \
                                .attrs['placeholder'] = 'Vendor\'s Part #?'
        self.fields['price'].widget.attrs['placeholder'] = 'Price?'

@permission_required(perms.VIEW)
def part_list(request):
    return Part_List_View.as_view()(request)

@permission_required(perms.VIEW)
def part_detail(request, pk):
    return Part_Detail_View.as_view()(request, pk=pk)

@permission_required(perms.VIEW)
def part_bom_detail(request, pk):
    # Part detail already includes BOM:
    return Part_Detail_View.as_view()(request, pk=pk)

class Assembly_Item_Edit_Formset(BaseInlineFormSet):
    def __init__(self, *args, **kwargs):
        super(Assembly_Item_Edit_Formset, self).__init__(*args, **kwargs)
        self.helper = crispy_inline_form_helper()

    def add_fields(self, form, index):
        super(Assembly_Item_Edit_Formset, self).add_fields(form, index)
        ndict = OrderedDict()
        # first, copy all hidden fields:
        for f in form.fields:
            if form.fields[f].widget.is_hidden:
                ndict[f] = form.fields[f]
        # second, rearrange the order of the visible fields and add
        # virtual (calculated) fields:
        if 'qty' in form.fields:
            ndict['qty'] = form.fields['qty']
            ndict['qty'].widget.attrs['size'] = 6
            ndict['qty'].widget.attrs['placeholder'] = 'Quantity?'
            ndict['qty'].help_text = ''
        ndict['comp'] = form.fields['comp']
        ndict['comp'].widget.attrs['size'] = 9
        ndict['comp'].help_text = ''
        ndict['mfg_pn'] = forms.CharField(label='Manufacturer &amp; Part #')
        ndict['mfg_pn'].widget.attrs['readonly'] = 'True'
        ndict['mfg_pn'].widget.attrs['tabindex'] = -1
        ndict['mfg_pn'].widget.attrs['size'] = 31
        ndict['refdes'] = form.fields['refdes']
        ndict['refdes'].widget.attrs['rows'] = 2
        ndict['refdes'].help_text = ''
        ndict[DELETION_FIELD_NAME] = form.fields[DELETION_FIELD_NAME]

        form.fields = ndict

    def initialize_virtual_fields(self):
        # setup initial data for calculated (virtual) fields:
        for form in self:
            init = form.initial
            if 'comp' in init:
                p = get_object_or_404(Part, pk=init['comp'])
                init['mfg_pn'] = p.mfg + ' ' + p.mfg_pn
            if 'refdes' in init:
                refdes_list = re.findall(r'([^, ]+)', init['refdes'])
                init['refdes'] = ' '.join(refdes_list)

    def clean(self):
        super(Assembly_Item_Edit_Formset, self).clean()

        for form in self.forms:
            if 'comp' in form.cleaned_data:
                comp = form.cleaned_data['comp']
                if comp.id == self.instance.id:
                    errors = form._errors.setdefault('comp', ErrorList())
                    errors.append('BOM may contain the assembly-part '
                                  'itself.')

@transaction.atomic
@permission_required(perms.EDIT)
def part_bom_edit(request, pk):
    part = get_object_or_404(Part, pk=pk)
    my_url = reverse('epic:part_bom_edit', kwargs={'pk': pk})
    prev_url = reverse('epic:part_detail', kwargs={'pk': pk})

    widgets = {
        'comp': AutocompleteTextWidget('Part_Autocomplete'),
    }
    BOM_Formset = inlineformset_factory(Part, Assembly_Item,
                                        formset=Assembly_Item_Edit_Formset,
                                        fk_name='assy', widgets=widgets,
                                        fields= [ 'comp', 'qty', 'refdes' ])
    if request.method == 'POST':
        if 'cancel' in request.POST:
            return HttpResponseRedirect(prev_url)

        bom_form = BOM_Formset(request.POST, instance=part)
        if bom_form.is_valid():
            part.last_bom_mod_type = Part.LAST_MOD_TYPE_USER
            part.last_bom_mod_name = request.user.get_username()
            part.save()

            bom_form.save()
            if 'save-and-done' in request.POST:
                return HttpResponseRedirect(prev_url)
            return HttpResponseRedirect(my_url)
    else:
        bom_form = BOM_Formset(instance=part)

    bom_form.initialize_virtual_fields()
    crumb = breadcrumb(my_url)
    return render(request, 'epic/bom_edit.html',
                  {
                      'pk': pk,
                      'bom_form': bom_form,
                      'breadcrumb': crumb,
                  })

@transaction.atomic
@permission_required(perms.EDIT)
def part_edit(request, pk):
    def prev_url(pk):
        if pk is None:
            return reverse('epic:part_list')
        return reverse('epic:part_detail', kwargs={'pk': pk})

    if pk is None:
        part = None
        my_url = reverse('epic:part_add')
    else:
        part = get_object_or_404(Part, pk=pk)
        my_url = reverse('epic:part_edit', kwargs={'pk': pk})

    Vendor_Part_Formset = inlineformset_factory(Part, Vendor_Part,
                                                form=Vendor_Part_Edit_Form,
                                            formset=Vendor_Part_Edit_Formset)

    if request.method == 'POST':
        if 'cancel' in request.POST:
            return HttpResponseRedirect(prev_url(pk))

        part_form = Part_Edit_Form(request.POST, instance=part)
        vendor_form = Vendor_Part_Formset(request.POST, instance=part)
        if part_form.is_valid() and vendor_form.is_valid():
            new_part = part_form.save()
            pk = new_part.pk

            new_part = get_object_or_404(Part, pk=pk)
            new_vendor_form = Vendor_Part_Formset(request.POST,
                                                  instance=new_part)
            new_vendor_form.is_valid()	# re-create cleaned data for .save()
            new_vendor_form.save()
            return HttpResponseRedirect(prev_url(pk))
    else:
        part_form = Part_Edit_Form(instance=part)
        vendor_form = Vendor_Part_Formset(instance=part)

    crumb = breadcrumb(my_url)
    return render(request, 'epic/part_edit.html',
                  {
                      'pk': pk,
                      'part_form': part_form,
                      'breadcrumb': crumb,
                      'vendor_form': vendor_form
                  })

@permission_required(perms.EDIT)
def part_add(request):
    return part_edit(request, None)

def here_are_all_or_some(item_name_prefix, suffix_singular, suffix_plural,
                         items, max_items=5):
    msg = 'Here '
    msg += 'is' if len(items) == 1 else 'are'
    msg += '' if len(items) <= max_items else ' some of'
    msg += ' the ' + item_name_prefix \
           + (suffix_singular if len(items) == 1 else suffix_plural) + ': '
    msg += ', '.join(items[:max_items])
    if len(items) > max_items:
        msg += ', &hellip;'
    else:
        msg += '.'
    return msg

def get_parent_url(url):
    m = re.match(r'(.*/)[^/]+/?$', url)
    if not m:
        raise Http404
    return m.group(1)

@permission_required(perms.EDIT)
def part_delete(request, pk):
    part = get_object_or_404(Part, pk=pk)
    my_url = reverse('epic:part_delete', kwargs={'pk': pk})
    parent_url = get_parent_url(my_url)

    # Check if the part is being referenced:
    #	1) In any Delta
    #	2) As a component in an assembly
    #	3) In any Line_Item
    # If so, refuse to delete, as that could throw everything off balance.
    # Note that it is OK if the Part is an Assembly_Item itself
    # or if a Vendor_Part refers to it (as long as the above conditions are
    # satisfied).  Those will be deleted automatically and there are no
    # negative effects in doing so.
    txtns  = Transaction.objects.filter(delta__part_id=part.id)
    txtns |= Transaction.objects.filter(line_item__part_id=part.id)
    txtns  = txtns.distinct().order_by('-id')
    assys  = Part.objects.filter(assembly_item_part__comp_id=part.id) \
                         .distinct()
    if txtns.exists() or assys.exists():
        messages = [ 'Sorry, part %s cannot be deleted because ' \
                     % part.html_link()]
        sep = ''
        if txtns.exists():
            messages[0] += '%d transaction%s refer%s to it' \
                           % (len(txtns), '' if len(txtns) == 1 else 's',
                              's' if len(txtns) == 1 else '')
            messages.append(here_are_all_or_some('transaction', '', 's',
                                                 [ txtn.html_link() \
                                                   for txtn in txtns ]))
            sep  = ' and '
        if assys.exists():
            messages[0] += '%s%d assembl%s use%s it' % \
                           (sep, len(assys),
                            'y' if len(assys) == 1 else 'ies',
                            's' if len(assys) == 1 else '')
            messages.append(here_are_all_or_some('assembl', 'y', 'ies',
                                                 [ assy.html_link() \
                                                   for assy in assys ]))
        messages[0] += '.'

        return render(request, 'epic/delete_error.html',
                      {
                          'breadcrumb': breadcrumb(my_url),
                          'parent_url': parent_url,
                          'messages': messages
                      })
    part.delete()
    return HttpResponseRedirect(get_parent_url(parent_url))

class Part_List_View(generic.ListView):
    model = Part
    paginate_by = 100	# this makes the the paging match the 3rd digit of PN

    def get_queryset(self):
        return super(Part_List_View, self).get_queryset().order_by('id')

    def get_context_data(self, **kwargs):
        context = super(Part_List_View, self).get_context_data(**kwargs)

        context['breadcrumb'] = breadcrumb(reverse('epic:part_list'))
        context['list_pager'] = html_list_pager(context['page_obj'],
                                                page_name=lambda pg: pg - 1)
        return context

class Part_Detail_View(generic.DetailView):
    model = Part
    fields = [ 'mfg', 'mfg_pn', 'val', 'descr', 'footprint', 'mounting',
               'target_price', 'overage', 'spq', 'lead_time', 'status' ]

    def get_history_filter(self):
        if 'ht' in self.request.GET:
            return self.request.GET['ht']
        return 'truncated'

    def get_context_data(self, **kwargs):

        def map_fields(name, value, verbose_name):
            if name != 'mfg_pn':
                return verbose_name, value
            args = {
                'mfg'   : self.object.mfg,
                'mfg_pn': self.object.mfg_pn
            }
            fname = '{mfg} {mfg_pn}'.format(**args)
            while True:
                ds_path = os.path.join(cfg.EPIC_DATASHEET_DIR, '%s.pdf' % fname)
                res = finders.find(ds_path)
                if finders.find(ds_path) is not None:
                    break
                # try after stripping off the stuff after the last dash
                # (e.g., -REEL7)
                m = re.match(r'(.*)[-/][^-/]+$', fname)
                if not m:
                    ds_path = None
                    break
                fname = m.group(1)
            ds_link = ''
            if ds_path is not None:
                ds_link = ('&ensp;<a href="%s" target="datasheet">'
                           '<i class="glyphicon glyphicon-book"></i></a>') \
                    % (os.path.join (cfg.STATIC_URL, ds_path))

            url = 'https://octopart.com/partsearch#!?q={mfg}%20{mfg_pn}' \
                .format(**args)
            return verbose_name, \
                ('%s%s&ensp;<a href="%s" target="part_lookup">'
                 '<img src="%s" alt=''></a>'
                 % (value, ds_link, url, os.path.join(cfg.STATIC_URL,
                                                      'epic/img/octopart.ico')))

        assemblies = get_assemblies_using_part(self.object)
        part_users = []
        for assy in assemblies:
            name = get_assembly_name(assy)
            url = reverse('epic:part_detail', kwargs={'pk': assy.id})
            part_users.append('<a href="%s">%s</a>' % (url, name))

        fields = get_model_fields(self.object, self.fields, map_fields)

        subs = self.object.equivalent_parts()
        subs.remove(self.object)
        subs = sorted(subs, key = lambda x: x.id)
        subs = html_list_or_none(map(lambda s: html_part_link(s), subs))

        fields.append([ 'Substitutes', subs ])

        vendor_parts = Vendor_Part.objects.filter(part_id=self.object.id)

        bom = [ ]
        total_target_cost = 0
        for item in self.object.assembly_items():
            best_part = item.comp.best_part()
            refdes_list = item.refdes.split(',')
            total_target_cost += item.qty * best_part.target_price
            avg_cost = best_part.avg_cost()
            bom.append({
                'best_part':		best_part,
                'assy_item':		item,
                'refdes_list':		refdes_list,
                'avg_cost':		avg_cost,
                'amount':		item.qty * avg_cost,
                'cumulative_cost':	0
            })
        bom.sort(key = lambda x: x['amount'], reverse=True)
        cumulative_cost = 0
        for row in bom:
            cumulative_cost += row['amount']
            row['cumulative_cost'] = cumulative_cost
        cost_summary = { 'part': {}, 'cmp': {}, 'tot': { 'tgt': 0, 'avg': 0 } }
        cost_summary['part']['tgt'] = self.object.target_price
        cost_summary['part']['avg'] = self.object.avg_cost()
        cost_summary['cmp']['avg'] = cumulative_cost
        cost_summary['cmp']['tgt'] = total_target_cost
        for kind in [ 'tgt', 'avg' ]:
            cost_summary['tot'][kind] += (cost_summary['part'][kind] +
                                          cost_summary['cmp'][kind])

        history = get_part_history(self.object)

        part_list = [ p.id for p in self.object.equivalent_parts() ]
        open_order_items = get_open_order_line_items() \
            .filter(part_id__in=part_list)

        pk = self.object.id
        ht = self.get_history_filter()

        context = super(Part_Detail_View, self).get_context_data(**kwargs)
        context['breadcrumb'] = breadcrumb(reverse('epic:part_detail',
                                                   kwargs={'pk': pk}))
        context['fields'] = fields
        context['vendor_parts'] = vendor_parts
        context['part_users'] = html_list_or_none(part_users)
        context['bom'] = bom
        context['cost_summary'] = cost_summary
        context['page_nav'] = html_page_nav('epic:part_detail', Part, pk)
        context['history'] = part_history_summary(self.object, history,
                                                  ht == 'full')
        context['ht'] = ht
        context['open_order_items'] = open_order_items
        return context

class Warehouse_Edit_Form(ModelForm):
    def __init__(self, *args, **kwargs):
        super(Warehouse_Edit_Form, self).__init__(*args, **kwargs)
        self.helper = crispy_form_helper()

    class Meta:
        model = Warehouse
        fields = [ 'name', 'address' ]

class Warehouse_List_View(generic.ListView):
    model = Warehouse
    paginate_by = 50

    def get_queryset(self):
        return super(Warehouse_List_View, self).get_queryset() \
                                               .order_by('name')

    def get_context_data(self, **kwargs):
        crumb = breadcrumb(reverse('epic:warehouse_list'))
        context = super(Warehouse_List_View, self).get_context_data(**kwargs)
        context['breadcrumb'] = crumb
        context['list_pager'] = html_list_pager(context['page_obj'])
        return context

class Warehouse_Detail_View(generic.DetailView):
    model = Warehouse
    fields = [ 'name', 'address' ]

    def get_context_data(self, **kwargs):
        pk = self.object.id

        inventories = self.object.inventories().order_by('-ts')
        paginator = Paginator(inventories, 50)
        page = self.request.GET.get('ipg')
        try:
            inventories = paginator.page(page)
        except PageNotAnInteger:
            inventories = paginator.page(1)
        except EmptyPage:
            inventories = paginator.page(paginator.num_pages)

        txtns  = Transaction.objects.filter(warehouse_id=pk)
        txtns |= Transaction.objects.filter(shipment__from_warehouse_id=pk)
        txtns = txtns.distinct().order_by('-ts')
        paginator = Paginator(txtns, 50)
        page = self.request.GET.get('tpg')
        try:
            txtns = paginator.page(page)
        except PageNotAnInteger:
            txtns = paginator.page(1)
        except EmptyPage:
            txtns = paginator.page(paginator.num_pages)

        context = super(Warehouse_Detail_View, self).get_context_data(**kwargs)
        # we don't want to depend on having
        # django.core.context_processors.request enabled:
        context['request'] = self.request
        context['fields'] = get_model_fields(self.object, self.fields)
        context['breadcrumb'] = breadcrumb(reverse('epic:warehouse_detail',
                                                   kwargs={'pk': pk}))
        context['inventories'] = inventories
        context['inventories_pager'] = html_list_pager(inventories, key='ipg')
        context['txtns'] = txtns
        context['txtns_pager'] = html_list_pager(txtns, key='tpg')
        return context

class Vendor_List_View(generic.ListView):
    model = Vendor
    paginate_by = 50

    def get_queryset(self):
        return super(Vendor_List_View, self).get_queryset().order_by('name')

    def get_context_data(self, **kwargs):
        crumb = breadcrumb(reverse('epic:vendor_list'))
        context = super(Vendor_List_View, self).get_context_data(**kwargs)
        context['breadcrumb'] = crumb
        context['list_pager'] = html_list_pager(context['page_obj'])
        return context

class Vendor_Detail_View(generic.DetailView):
    model = Vendor
    fields = [ 'name', 'search_url' ]

    def get_context_data(self, **kwargs):
        pk = self.object.id

        vps = Vendor_Part.objects.filter(vendor_id=pk).order_by('id')
        paginator = Paginator(vps, 50)
        page = self.request.GET.get('vpg')
        try:
            vps = paginator.page(page)
        except PageNotAnInteger:
            vps = paginator.page(1)
        except EmptyPage:
            vps = paginator.page(paginator.num_pages)

        txtns  = Transaction.objects.filter(order__vendor__id=pk)
        txtns |= Transaction.objects.filter(shipment__ordr__vendor_id=pk)
        txtns  = txtns.distinct().order_by('-ts')
        paginator = Paginator(txtns, 50)
        page = self.request.GET.get('tpg')
        try:
            txtns = paginator.page(page)
        except PageNotAnInteger:
            txtns = paginator.page(1)
        except EmptyPage:
            txtns = paginator.page(paginator.num_pages)

        context = super(Vendor_Detail_View, self).get_context_data(**kwargs)
        # we don't want to depend on having
        # django.core.context_processors.request enabled:
        context['request'] = self.request
        context['vps'] = vps
        context['vps_pager'] = html_list_pager(vps, key='vpg')
        context['txtns'] = txtns
        context['txtns_pager'] = html_list_pager(txtns, key='tpg')
        context['breadcrumb'] = breadcrumb(reverse('epic:vendor_detail',
                                                   kwargs={'pk': pk}))
        context['fields'] = get_model_fields(self.object, self.fields)
        return context

class Vendor_Edit_Form(ModelForm):
    def __init__(self, *args, **kwargs):
        super(Vendor_Edit_Form, self).__init__(*args, **kwargs)
        self.helper = crispy_form_helper()

    class Meta:
        model = Vendor
        fields = [ 'name', 'search_url' ]

@permission_required(perms.VIEW)
def vendor_list(request):
    return Vendor_List_View.as_view()(request)

@permission_required(perms.VIEW)
def vendor_detail(request, pk):
    return Vendor_Detail_View.as_view()(request, pk=pk)

@transaction.atomic
@permission_required(perms.EDIT)
def vendor_edit(request, pk):
    def prev_url(pk):
        if pk is None:
            return reverse('epic:vendor_list')
        return reverse('epic:vendor_detail', kwargs={'pk': pk})

    if pk is None:
        vendor = None
        my_url = reverse('epic:vendor_add')
    else:
        vendor = get_object_or_404(Vendor, pk=pk)
        my_url = reverse('epic:vendor_edit', kwargs={'pk': pk})

    if request.method == 'POST':
        if 'cancel' in request.POST:
            return HttpResponseRedirect(prev_url(pk))

        form = Vendor_Edit_Form(request.POST, instance=vendor)
        if form.is_valid():
            new_form = form.save()
            return HttpResponseRedirect(prev_url(new_form.pk))
    else:
        form = Vendor_Edit_Form(instance=vendor)

    data_lists = ''
    crumb = breadcrumb(my_url)
    return render(request, 'epic/vendor_edit.html',
                  {
                      'form': form,
                      'pk': pk,
                      'breadcrumb': crumb,
                      'data_lists': data_lists
                  })

@permission_required(perms.EDIT)
def vendor_add(request):
    return vendor_edit(request, None)

@permission_required(perms.EDIT)
def vendor_delete(request, pk):
    vendor = get_object_or_404(Vendor, pk=pk)
    my_url = reverse('epic:vendor_delete', kwargs={'pk': pk})
    parent_url = get_parent_url(my_url)

    # Check if the vendor is being referenced:
    #	1) In any Order
    #	2) In any Vendor_Part
    # If so, refuse to delete, as that could throw everything off balance.
    orders  = Order.objects.filter(vendor_id=vendor.id).order_by('-id')
    vps = Vendor_Part.objects.filter(vendor_id=vendor.id).order_by('id')
    if orders.exists() or vps.exists():
        messages = [ 'Sorry, vendor %s cannot be deleted because ' \
                     % vendor.html_link() ]
        sep = ''
        if orders.exists():
            messages[0] += '%d order%s refer%s to it' \
                           % (len(orders), '' if len(orders) == 1 else 's',
                              's' if len(orders) == 1 else '')
            messages.append(here_are_all_or_some('order', '', 's',
                                                 [ order.html_link() \
                                                   for order in orders ]))
            sep  = ' and '
        if vps.exists():
            messages[0] += '%s%d vendor part%s refer%s to it' % \
                           (sep, len(vps), '' if len(vps) == 1 else 's',
                            's' if len(vps) == 1 else '')
            messages.append(here_are_all_or_some('part', '', 's',
                                                 [ vp.part.html_link() \
                                                   for vp in vps ]))
        messages[0] += '.'
        return render(request, 'epic/delete_error.html',
                      {
                          'breadcrumb': breadcrumb(my_url),
                          'parent_url': parent_url,
                          'messages': messages
                      })
    vendor.delete()
    return HttpResponseRedirect(get_parent_url(parent_url))

@permission_required(perms.VIEW)
def warehouse_stock(request, pk=None):
    context = {}

    title = 'Current Stock '
    if pk is None:
        warehouse = None
        title += 'for all warehouses together'
        my_url = reverse('epic:warehouse_stock_all')
    else:
        warehouse = get_object_or_404(Warehouse, pk=pk)
        title += 'at warehouse %s' % warehouse.name
        my_url = reverse('epic:warehouse_stock', kwargs={'pk': pk})

    if request.method == 'POST':
        return HttpResponseRedirect(my_url + '?as_of_date=%s' \
                                    % request.POST.get('as_of_date'))
    elif 'as_of_date' in request.GET:
        as_of_date = datetime.strptime(request.GET['as_of_date'],
                                       '%Y-%m-%d').date()
    else:
        as_of_date = date.today()

    title += ' as of %s' % defaultfilters.date(as_of_date, 'Y-m-d')

    stock, stock_total = get_stock_summary(warehouse, as_of_date=as_of_date)

    context['pk'] = pk
    context['breadcrumb'] = breadcrumb(my_url)
    context['title'] = title
    context['stock'] = stock
    context['stock_total'] = stock_total
    context['as_of_date'] = as_of_date
    return render(request, 'epic/warehouse_stock.html', context)

@permission_required(perms.VIEW)
def warehouse_stock_all(request, ):
    return warehouse_stock(request, pk=None)

@permission_required(perms.VIEW)
def warehouse_list(request):
    return Warehouse_List_View.as_view()(request)

@permission_required(perms.VIEW)
def warehouse_detail(request, pk):
    return Warehouse_Detail_View.as_view()(request, pk=pk)

@permission_required(perms.VIEW)
def warehouse_inventory(request, pk):
    # dummy to make breadcrumbs work...
    return warehouse_detail(request, pk)

@transaction.atomic
@permission_required(perms.EDIT)
def warehouse_edit(request, pk):
    def prev_url(pk):
        if pk is None:
            return reverse('epic:warehouse_list')
        return reverse('epic:warehouse_detail', kwargs={'pk': pk})

    if pk is None:
        warehouse = None
        my_url = reverse('epic:warehouse_add')
    else:
        warehouse = get_object_or_404(Warehouse, pk=pk)
        my_url = reverse('epic:warehouse_edit', kwargs={'pk': pk})

    if request.method == 'POST':
        if 'cancel' in request.POST:
            return HttpResponseRedirect(prev_url(pk))
        form = Warehouse_Edit_Form(request.POST, instance=warehouse)
        if form.is_valid():
            new_form = form.save()
            return HttpResponseRedirect(prev_url(new_form.pk))
    else:
        form = Warehouse_Edit_Form(instance=warehouse)

    crumb = breadcrumb(my_url)
    return render(request, 'epic/warehouse_edit.html', { 'form': form,
                                                         'pk': pk,
                                                         'breadcrumb': crumb })

@permission_required(perms.EDIT)
def warehouse_add(request):
    return warehouse_edit(request, None)

@permission_required(perms.EDIT)
def warehouse_delete(request, pk):
    warehouse = get_object_or_404(Warehouse, pk=pk)
    my_url = reverse('epic:warehouse_delete', kwargs={'pk': pk})
    parent_url = get_parent_url(my_url)

    # Check if the warehouse is being referenced:
    #	1) In any Delta
    #	2) In any Transaction
    # If so, refuse to delete, as that could throw everything off balance.
    # Note: we don't check explicitly if there are any shipments where
    # Shipment.from_warehouse is set: if the shipment has no deltas, it's
    # ok to delete it (which will happen automatically) and if it does,
    # then it'd show up as a delta and line-item.
    txtns  = Transaction.objects.filter(delta__warehouse_id=warehouse.id)
    txtns |= Transaction.objects.filter(warehouse_id=warehouse.id)
    txtns  = txtns.distinct().order_by('-id')
    if txtns.exists():
        messages = [ 'Sorry, warehouse %s cannot be deleted because ' \
                     % warehouse.html_link()]
        sep = ''
        if txtns.exists():
            messages[0] += '%d transaction%s refer%s to it' \
                           %(len(txtns), '' if len(txtns) == 1 else 's',
                             's' if len(txtns) == 1 else '')
            messages.append(here_are_all_or_some('transaction', '', 's',
                                                 [ txtn.html_link() \
                                                   for txtn in txtns ]))
        messages[0] += '.'
        return render(request, 'epic/delete_error.html',
                      {
                          'breadcrumb': breadcrumb(my_url),
                          'parent_url': parent_url,
                          'messages': messages
                      })
    warehouse.delete()
    return HttpResponseRedirect(get_parent_url(parent_url))

@permission_required(perms.EDIT)
def warehouse_add_shipment(request, pk):
    return ship_edit(request, None, from_warehouse_id=pk)

@permission_required(perms.EDIT)
def warehouse_add_inventory(request, pk):
    return inventory_edit(request, None, warehouse=pk)

class Warehouse_Inv_Detail_View(generic.DetailView):
    model = Inventory
    template_name = 'epic/inventory_detail.html'

    def __init__(self, *args, **kwargs):
        super(Warehouse_Inv_Detail_View, self).__init__(*args, **kwargs)

    def get_context_data(self, **kwargs):
        pk = self.object.id
        warehouse_id = int(self.kwargs['warehouse'])

        if warehouse_id != self.object.warehouse_id:
            raise Http404

        inventory_items, total_value_change, has_relative_deltas \
            = get_inventory_summary(warehouse_id, self.object)

        context = super(Warehouse_Inv_Detail_View, self) \
            .get_context_data(**kwargs)
        context['pk'] = pk
        context['breadcrumb'] = breadcrumb(
            reverse('epic:warehouse_inventory_detail',
                    kwargs={'warehouse': warehouse_id, 'pk': pk}))
        context['warehouse'] = warehouse_id
        context['inventory_items'] = inventory_items
        context['total_value_change'] = total_value_change
        context['has_relative_deltas'] = has_relative_deltas
        return context

@permission_required(perms.VIEW)
def warehouse_inv_detail(request, warehouse, pk):
    return Warehouse_Inv_Detail_View.as_view()(request,
                                               warehouse=warehouse, pk=pk)

class Inventory_Edit_Form(autocomplete_light.ModelForm):
    def __init__(self, fixed, *args, **kwargs):
        super(Inventory_Edit_Form, self).__init__(*args, **kwargs)
        self.fixed = fixed
        self.helper = crispy_form_helper()
        self.has_assemblies = False
        self.fields['ts'].label = 'Created on'

    class Meta:
        model = Inventory
        fields = [ 'ts', 'notes' ]

class Inventory_Edit_Form(autocomplete_light.ModelForm):
    def __init__(self, *args, **kwargs):
        super(Inventory_Edit_Form, self).__init__(*args, **kwargs)
        self.helper = crispy_form_helper()
        self.has_assemblies = False
        self.fields['ts'].label = 'Inventory Date'
        self.fields['ts'].input_formats = [ '%Y-%m-%d %I:%M %p']

    class Meta:
        model = Inventory
        fields = [ 'ts', 'notes' ]
        autocomplete_names = { 'part': 'Part_Autocomplete' }
        widgets = {
            'ts': DateTimePicker(
                options={
                    'format': 'YYYY-MM-DD hh:mm a', 'pickTime': True
                })
        }

@transaction.atomic
@permission_required(perms.EDIT)
def inventory_edit(request, pk, warehouse):
    def prev_url(pk):
        if pk is None:
            return reverse('epic:warehouse_list')
        return reverse('epic:warehouse_inventory_detail',
                       kwargs={'warehouse': warehouse, 'pk': pk})

    if pk is None:
        inventory = None
        my_url = reverse('epic:warehouse_add_inventory',
                          kwargs={'pk': warehouse})
    else:
        inventory = get_object_or_404(Inventory, pk=pk)
        if int(warehouse) != inventory.warehouse_id:
            raise Http404
        for d in Delta.objects.filter(txtn_id=pk):
            if not d.is_absolute:
                raise Http404
        my_url = reverse('epic:warehouse_inventory_edit',
                         kwargs={'warehouse': warehouse, 'pk': pk})

    widgets = {
        'part': AutocompleteTextWidget('Part_Autocomplete'),
    }
    Delta_Formset = inlineformset_factory(Transaction, Delta,
                                          formset=Line_Item_Edit_Formset,
                                          fields=[ 'part', 'adj' ],
                                          widgets=widgets)
    if request.method == 'POST':
        if 'cancel' in request.POST:
            return HttpResponseRedirect(prev_url(pk))

        inv_form = Inventory_Edit_Form(request.POST, instance=inventory)
        item_form = Delta_Formset(request.POST, instance=inventory)

        if inv_form.is_valid() and item_form.is_valid():
            pk = item_form.save_with_transaction(request, Delta_Formset,
                                                 inv_form, item_form,
                                                 int(warehouse))
            if 'save-and-done' in request.POST:
                return HttpResponseRedirect(prev_url(pk))
            return HttpResponseRedirect(
                reverse('epic:warehouse_inventory_edit',
                        kwargs={'warehouse': warehouse, 'pk': pk}))
    else:
        inv_form = Inventory_Edit_Form(instance=inventory)
        if inventory is None:
            inv_form.fields['ts'].initial = timezone.now()
        item_form = Delta_Formset(instance=inventory)

    item_form.initialize_virtual_fields(None)

    return render(request, 'epic/inventory_edit.html',
                  {
                      'pk': pk,
                      'breadcrumb': breadcrumb(my_url),
                      'warehouse': warehouse,
                      'inv_form': inv_form,
                      'item_form': item_form
                  })

@permission_required(perms.EDIT)
def inventory_delete(request, pk, warehouse):
    inv = get_object_or_404(Inventory, pk=pk)
    if inv.warehouse_id != int(warehouse):
        raise Http404
    my_url = reverse('epic:warehouse_inventory_delete',
                     kwargs={'warehouse': warehouse, 'pk': pk})
    parent_url = get_parent_url(my_url)

    # Inventories are easy: nothing else refers to them so we can always delete
    # them.
    inv.delete()
    return HttpResponseRedirect(get_parent_url(get_parent_url(parent_url)))

class Order_List_View(generic.ListView):
    model = Order
    paginate_by = 50

    def get_status_filter(self):
        if self.request.method == 'GET' and('st' in self.request.GET):
            return self.request.GET['st']
        return 'open'

    def get_queryset(self):
        qs = Order.objects.order_by('-expected_arrival_date')
        st = self.get_status_filter()
        if st == 'open':
            qs = qs.filter(status=Order.STATUS_OPEN)
        elif st == 'closed':
            qs = qs.filter(status=Order.STATUS_CLOSED)
        return qs

    def get_context_data(self, **kwargs):
        st = self.get_status_filter()
        query_string = ''
        if st != 'open':
            query_string = 'st=%s' % st

        context = super(Order_List_View, self).get_context_data(**kwargs)
        context['breadcrumb'] = breadcrumb(reverse('epic:order_list'))
        context['list_pager'] = html_list_pager(context['page_obj'],
                                                queries=[ query_string ])
        context['st'] = st
        return context

class Order_Detail_View(generic.DetailView):
    model = Order
    fields = [ 'status', 'ts', 'expected_arrival_date' ]

    def get_context_data(self, **kwargs):
        def map_fields(name, value, verbose_name):
            if name == 'ts':
                # Python 3 includes microsecs; Python 2 does not:
                try:
                    ts = datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f+00:00')
                except ValueError:
                    ts = datetime.strptime(value, '%Y-%m-%d %H:%M:%S+00:00')
                return 'Order Date', \
                    defaultfilters.date(ts.replace(tzinfo=timezone.utc),
                                        'Y-m-d')
            return verbose_name, value

        order = self.object

        shipments = []
        for s in Shipment.objects.filter(ordr_id=order.id):
            shipments.append(s.html_link())

        order_items = Line_Item.objects.filter(txtn_id=order.id)

        pk = order.id

        context = super(Order_Detail_View, self).get_context_data(**kwargs)
        context['breadcrumb'] = breadcrumb(reverse('epic:order_detail',
                                                   kwargs={'pk': pk}))
        context['fields'] = get_model_fields(order, self.fields, map_fields)
        context['page_nav'] = html_page_nav('epic:order_detail', Order, pk)
        context['order'] = order
        context['order_items'] = order_items
        context['shipments'] = html_list_or_none(shipments)
        return context

class Transaction_Edit_Form(ModelForm):
    def __init__(self, *args, **kwargs):
        super(Transaction_Edit_Form, self).__init__(*args, **kwargs)
        self.helper = crispy_form_helper()

    class Meta:
        model = Transaction
        fields = [ 'notes' ]

class Order_Edit_Form(autocomplete_light.ModelForm):
    def __init__(self, *args, **kwargs):
        super(Order_Edit_Form, self).__init__(*args, **kwargs)
        self.helper = crispy_form_helper()
        self.has_assemblies = False
        self.fields['ts'].label = 'Order Date'
        self.fields['warehouse'].label = 'Ship-To Warehouse'

    def clean(self):
        cleaned_data = super(Order_Edit_Form, self).clean()
        if self.has_assemblies:
            vendor = cleaned_data['vendor']
            try:
                warehouse = Warehouse.by_name(vendor.name)
            except ObjectDoesNotExist:
                raise forms.ValidationError(
                    {'vendor': ['Order contains assemblies but this '
                                'vendor has no warehouse of the same name.  '
                                'Please change vendor or create a warehouse '
                                'for this vendor.']})

    class Meta:
        model = Order
        fields = [ 'ts', 'vendor', 'warehouse', 'expected_arrival_date',
                   'notes' ]
        autocomplete_names = { 'part': 'Part_Autocomplete' }
        widgets = {
            'ts': DateTimePicker(options={'format': 'YYYY-MM-DD',
                                          'pickTime': False}),
            'expected_arrival_date': DateTimePicker(options={'format':
                                                             'YYYY-MM-DD',
                                                             'pickTime': False})
        }

class Line_Item_Edit_Formset(BaseInlineFormSet):
    def __init__(self, *args, **kwargs):
        super(Line_Item_Edit_Formset, self).__init__(*args, **kwargs)
        self.helper = crispy_inline_form_helper()

    def add_fields(self, form, index):
        super(Line_Item_Edit_Formset, self).add_fields(form, index)
        ndict = OrderedDict()
        # first, copy all hidden fields:
        for f in form.fields:
            if form.fields[f].widget.is_hidden:
                ndict[f] = form.fields[f]
        # second, rearrange the order of the visible fields and add
        # virtual (calculated) fields:
        if 'qty' in form.fields:
            ndict['qty'] = form.fields['qty']
            ndict['qty'].widget.attrs['size'] = 6
            ndict['qty'].widget.attrs['placeholder'] = 'Quantity?'
            ndict['qty'].help_text = ''
        elif 'adj' in form.fields:
            ndict['adj'] = form.fields['adj']
            ndict['adj'].widget.attrs['size'] = 6
            ndict['adj'].widget.attrs['placeholder'] = 'Quantity?'
            ndict['adj'].label = 'Qty'
            ndict['adj'].help_text = ''
        ndict['part'] = form.fields['part']
        ndict['part'].widget.attrs['size'] = 9
        ndict['part'].help_text = ''
        ndict['mfg_pn'] = forms.CharField(label='Manufacturer &amp; Part #')
        ndict['mfg_pn'].widget.attrs['readonly'] = 'True'
        ndict['mfg_pn'].widget.attrs['tabindex'] = -1
        ndict['mfg_pn'].widget.attrs['size'] = 31
        if 'qty' in form.fields:
            ndict['vendor_pn'] = forms.CharField(label='Vendor\'s Part #')
            ndict['vendor_pn'].widget.attrs['readonly'] = 'True'
            ndict['vendor_pn'].widget \
                              .attrs['placeholder'] = 'Vendors\'s Part #?'
            ndict['part_cost'] = forms.CharField(label='Price')
            ndict['part_cost'].widget.attrs['size'] = 8
            ndict['part_cost'].widget.attrs['placeholder'] = 'Price?'
        if 'line_cost' in form.fields:
            ndict['line_cost'] = form.fields['line_cost']
            ndict['line_cost'].widget.attrs['size'] = 8
            ndict['line_cost'].widget \
                              .attrs['placeholder'] = 'Line-item cost?'
            ndict['line_cost'].help_text = ''
        ndict[DELETION_FIELD_NAME] = form.fields[DELETION_FIELD_NAME]

        form.fields = ndict

    def clean(self):
        super(Line_Item_Edit_Formset, self).clean()

        parts_listed = {}
        line = 0
        for form in self.forms:
            line += 1
            if 'part' not in form.cleaned_data or \
               'part_cost' not in form.cleaned_data or \
               'vendor_pn' not in form.cleaned_data:
                continue
            part = form.cleaned_data['part']
            if part is None:
                continue
            if part in parts_listed:
                # same part is listed multiple times
                errors = form._errors.setdefault('part', ErrorList())
                errors.append('Line %d already lists this part.' \
                              % parts_listed[part])
                continue
            else:
                parts_listed[part] = line
            if form.instance.pk is None:
                # When adding a new line-item, make sure it's part # doesn't
                # conflict with one that's already existing.  If it doesn,
                # add error messages that tell the user how to fix the problem.
                # Since this runs
                # after django.db.models._perform_unique_checks,
                # we can't fix the problem automatically.
                qs = form.instance.__class__.objects \
                                            .filter(txtn_id=self.instance.id) \
                                            .filter(part_id=part.id)
                if qs.exists():
                    existing_pk = qs[0].pk
                    for other_form in self.forms:
                        if other_form == form:
                            continue
                        if other_form.instance.pk == existing_pk:
                            other_form.instance._set_pk_val(None)
                            errors = other_form._errors.setdefault('part',
                                                                   ErrorList())
                            errors.append('Please use this for part %s.' \
                                          % part)
                            errors = form._errors.setdefault('part',
                                                             ErrorList())
                            errors.append('Please use this for part %s.' \
                                          % other_form.instance.part)
                            break
            part_cost = Decimal(form.cleaned_data['part_cost'])
            vendor_pn = form.cleaned_data['vendor_pn']
            delete = form.cleaned_data[DELETION_FIELD_NAME]
            txtn = self.instance
            if not delete and hasattr(txtn, 'order') \
               and txtn.order.vendor_id is not None:
                # Update Vendor_Part number if necessary:
                vp = Vendor_Part.get(part_id=part.id,
                                     vendor_id=txtn.order.vendor_id)
                if vp is None:
                    nvp = Vendor_Part(part_id=part.id,
                                      vendor_id=txtn.order.vendor_id,
                                      vendor_pn=vendor_pn,
                                      price=part_cost)
                    try:
                        nvp.full_clean()
                        nvp.save()
                    except ValidationError as e:
                        errors = form._errors.setdefault('vendor_pn',
                                                         ErrorList())
                        for msg in e.message_dict[NON_FIELD_ERRORS]:
                            errors.append(msg)
                elif part_cost != vp.price:
                    vp.price = part_cost
                    vp.save()

                # Update target_price of part if it's higher than part_cost:
                if part_cost < part.target_price:
                    part.target_price = part_cost
                    part.save()

    def initialize_virtual_fields(self, vendor_id):
        # setup initial data for calculated (virtual) fields:
        for form in self:
            init = form.initial
            if 'part' in init:
                p = get_object_or_404(Part, pk=init['part'])
                init['mfg_pn'] = p.mfg + ' ' + p.mfg_pn
                if vendor_id is not None:
                    vp = Vendor_Part.get(part_id=init['part'],
                                         vendor_id=vendor_id)
                    if vp is not None:
                        init['vendor_pn'] = vp.vendor_pn
            if 'line_cost' in init and 'qty' in init:
                init['part_cost'] = part_cost(init['line_cost'], init['qty'])

    def save_with_transaction(self, request, formset_create,
                              txtn_form, item_form, warehouse_id=None):
        """Handle the common steps of saving the line-items and it's
        transaction.  txtn_form and item_form must have been validated
        before calling this function.

        """
        txtn = txtn_form.save(commit=False)
        if txtn.ts is None:
            txtn.ts = timezone.now()
        if txtn.warehouse_id is None:
            txtn.warehouse_id = warehouse_id
        txtn.save()

        new_item_form = formset_create(request.POST, instance=txtn)
        new_item_form.is_valid()	# re-create cleaned data for .save()

        new_line_items = new_item_form.save(commit=False)

        for item in new_item_form.deleted_objects:
            # With Django 1.7, formset.save (..., commit=False) does
            # not delete automatically.  We need to delete before saving
            # because otherwise we might end up with two entries of the
            # same part...
            item.delete()

        next_index = txtn.first_available_index()
        for item in new_line_items:
            if hasattr(item, 'index'):
                if item.index is None:
                    item.index = next_index
                    next_index += 1
            if hasattr(item, 'warehouse_id'):
                # initialize Delta fields that are not editable:
                item.warehouse_id = warehouse_id
                item.is_absolute = True
            item.save()
        txtn.finalize()
        return txtn.id

@permission_required(perms.VIEW)
def order_list(request):
    return Order_List_View.as_view()(request)

@permission_required(perms.VIEW)
def order_detail(request, pk):
    return Order_Detail_View.as_view()(request, pk=pk)

@transaction.atomic
@permission_required(perms.EDIT)
def order_edit(request, pk, order_initial=None, items_initial=None):
    def prev_url(pk):
        if pk is None:
            return reverse('epic:order_list')
        return reverse('epic:order_detail', kwargs={'pk': pk})

    if pk is None:
        order = None
        my_url = reverse('epic:order_add')
    else:
        order = get_object_or_404(Order, pk=pk)
        my_url = reverse('epic:order_edit', kwargs={'pk': pk})

    # See http://bit.ly/1wfFNKa:
    if order_initial is None:
        order_initial = {}
    if items_initial is None:
        items_initial = []

    widgets = {
        'part': AutocompleteTextWidget('Part_Autocomplete')
    }
    num_extra = len(items_initial) + 3
    Line_Item_Formset = inlineformset_factory(Order, Line_Item,
                                              formset=Line_Item_Edit_Formset,
                                              fields=[ 'part', 'qty',
                                                       'line_cost' ],
                                              widgets=widgets, extra=num_extra)

    if request.method == 'POST' and 'preset' not in request.POST:
        if 'cancel' in request.POST:
            return HttpResponseRedirect(prev_url(pk))

        order_form = Order_Edit_Form(request.POST, instance=order)
        item_form = Line_Item_Formset(request.POST, instance=order)

        #
        # First, do a partial validation of the order items so we can check
        # if any assemblies are being ordered.
        #
        if item_form.is_valid():
            for form in item_form.forms:
                if not 'part' in form.cleaned_data:
                    continue
                if form.cleaned_data['part'].assembly_items().exists():
                    order_form.has_assemblies = True
                    break

            if order_form.is_valid():
                pk = item_form.save_with_transaction(request,
                                                     Line_Item_Formset,
                                                     order_form, item_form)
                if 'save-and-done' in request.POST:
                    return HttpResponseRedirect(prev_url(pk))
                return HttpResponseRedirect(reverse('epic:order_edit',
                                                    kwargs={'pk': pk}))
    else:
        if order is None and len(order_initial) == 0:
            order_initial['ts'] = datetime.today()
        order_form = Order_Edit_Form(instance=order, initial=order_initial)
        item_form = Line_Item_Formset(instance=order, initial=items_initial)

    item_form.initialize_virtual_fields(order_form.initial.get('vendor'));

    crumb = breadcrumb(my_url)
    return render(request, 'epic/order_edit.html',
                  {
                      'order_form': order_form,
                      'item_form': item_form,
                      'pk': pk,
                      'breadcrumb': crumb,
                  })

@permission_required(perms.EDIT)
def order_add(request):
    order_initial = {}
    items_initial = []
    if request.method == 'POST' and 'preset' in request.POST:
        order_initial, \
            items_initial = get_initial_from_post(request, 'item_set',
                                                  Order, Line_Item)
    return order_edit(request, None, order_initial, items_initial)

@permission_required(perms.EDIT)
def order_delete(request, pk):
    order = get_object_or_404(Order, pk=pk)
    my_url = reverse('epic:order_delete', kwargs={'pk': pk})
    parent_url = get_parent_url(my_url)

    # Check if the order is being referenced:
    #	1) In any Shipment
    # If so, refuse to delete, as that could throw everything off balance.
    ships  = Shipment.objects.filter(ordr_id=order.id).order_by('-id')
    if ships.exists():
        messages = [ 'Sorry, order %s cannot be deleted because '
                     '%d shipment%s refer%s to it.'
                     % (order.html_link(),
                        len(ships), '' if len(ships) == 1 else 's',
                        's' if len(ships) == 1 else '') ]
        messages.append(here_are_all_or_some('shipment', '', 's',
                                             [ ship.transaction_ptr.html_link()
                                               for ship in ships ]))
        return render(request, 'epic/delete_error.html',
                      {
                          'breadcrumb': breadcrumb(my_url),
                          'parent_url': parent_url,
                          'messages': messages
                      })
    order.delete()
    return HttpResponseRedirect(get_parent_url(parent_url))

class Makeup_Order_Item:
    def __init__(self, src, part, qty):
        self.part = part
        self.amount = None
        if isinstance(src, Vendor):
            self.vendor_part = Vendor_Part.get(part.id, src.id)
        else:
            self.vendor_part = None
        self.set_qty(qty)

    def set_qty(self, qty):
        self.qty = qty
        if self.vendor_part:
            self.amount = self.qty * self.vendor_part.price

    def __repr__(self):
        rest = ''
        if self.vendor_part:
            rest =(',vendor_part=%s,amount=%s' % (repr(self.vendor_part),
                                                  self.amount))
        return 'Makeup_Order_Item(part=%s,qty=%s%s)' % (self.part,
                                                        self.qty, rest)

class Makeup_Order:
    def __init__(self, src, dst):
        self.src = src
        self.dst = dst
        self.items = []

    def add_part(self, part, qty):
        self.items.append(Makeup_Order_Item(self.src, part, qty))

    def finalize(self):
        for item in self.items:
            if isinstance(self.src, Vendor):
                spq = item.part.spq
                if spq > 1:
                    item.set_qty(int(math.ceil(float(item.qty) / spq)) * spq)
        self.items.sort(key=operator.attrgetter('part.id'))

    def __repr__(self):
        return 'Makeup_Order(%s->%s): %s' % \
           (self.src.name, self.dst.name, repr(self.items))

def add_to_makeup_orders(orders, for_warehouse, part, qty_short, inv):
    # first, see if we can ship parts from another warehouse:
    for w in inv.warehouse:
        if w == for_warehouse.id:
            continue
        for p in part.equivalent_parts():
            have = inv.get_qty(w, p.id)
            # only arrange for inter-warehouse shipment for this part if the
            # other warehouse has enough parts to satisfy the shortage or
            # it has at least standard-package-quantity pieces:
            if have >= qty_short or have >= part.spq:
                if have >= qty_short:
                    take = qty_short
                else:
                    take = have
                warehouse = get_object_or_404(Warehouse, pk=w)
                if warehouse not in orders:
                    orders[warehouse] = Makeup_Order(warehouse, for_warehouse)
                orders[warehouse].add_part(p, qty=take)
                qty_short -= take
                inv.adj_qty(w, p.id, -take)

    # second, order the parts from a vendor
    if qty_short > 0:
        vendor = part.best_vendor()
        if not vendor:
            return
        if vendor not in orders:
            orders[vendor] = Makeup_Order(vendor, for_warehouse)
        orders[vendor].add_part(part, qty_short)

@permission_required(perms.VIEW)
def order_check_stock(request, pk):
    def shortage_key(x):
        qty = x['warehouses'][0]['qty_remaining']
        if qty < 0:
            return qty
        return x['part'].id

    crumb = breadcrumb(reverse('epic:order_check_stock', kwargs={'pk': pk}))
    order = get_object_or_404(Order, pk=pk)
    if order.status != Order.STATUS_OPEN or \
       not order.assembly_line_items().exists():
        raise Http404	# order not open or has no line-items for assemblies
    open_orders = get_open_order_summary(last_order=order)

    inv = get_stock()

    my_summary_items = []
    for summary in open_orders:
        if summary['order'].id == order.id:
            my_summary_items = summary['items']
        for item in summary['items']:
            inv.apply_order_line_item(item)

    comp_status = []
    order_warehouse = Warehouse.by_name(order.vendor.name)
    max_lead_time = 0
    warehouses = [ order_warehouse.id ]
    max_assy = None
    for line_item in my_summary_items:
        assy_lead_time = line_item.part.lead_time
        if max_assy is None:
            max_assy = line_item.part
        for assy_item in line_item.part.assembly_items():
            if assy_lead_time + assy_item.comp.lead_time > max_lead_time:
                max_lead_time = assy_lead_time + assy_item.comp.lead_time
                max_assy = line_item.part
                max_comp = assy_item.comp
            part = assy_item.comp.best_part()
            if any(sts['part'] == part for sts in comp_status):
                continue	# we already checked this part
            comp_status.append({
                'part': part,
                'warehouses': [ ]
            })
            for w in inv.warehouse:
                for p in part.equivalent_parts():
                    if inv.get_qty(w, p.id) != 0:
                        try:
                            warehouses.index(w)
                        except:
                            warehouses.append(w)

    min_assy_arrival_date = date.today() + timedelta(days=7 * max_lead_time)

    needed_by = (order.expected_arrival_date
                 - timedelta(days=7 * max_assy.lead_time))
    tomorrow = date.today() + timedelta(days = 1)
    if needed_by < tomorrow:
        needed_by = tomorrow
    available_time = order.expected_arrival_date - date.today()
    available_time = math.ceil(available_time.days / 7)

    shortages = {}
    makeup_orders = {}
    for sts in comp_status:
        part = sts['part']
        for w in warehouses:
            total_qty = 0
            qty_breakdown = []
            for p in part.equivalent_parts():
                qty = inv.get_qty(w, p.id)
                total_qty += qty
                if qty != 0:
                    qty_breakdown.append({ 'part': p, 'qty' : qty })
            if total_qty < 0 and w == order_warehouse.id:
                shortages[part.id] = -total_qty
            warehouse_sts = { 'qty_remaining': total_qty }
            if len(qty_breakdown) > 1:
                warehouse_sts['qty_breakdown'] = qty_breakdown

            sts['warehouses'].append(warehouse_sts)
        if part.id in shortages:
            add_to_makeup_orders(makeup_orders, order_warehouse,
                                 part, shortages[part.id], inv)
    makeup_order_list = []
    for warehouse, makeup_order in makeup_orders.items():
        makeup_order.finalize()
        makeup_order_list.append(makeup_order)
    # list inter-warehouse shipments first:
    makeup_order_list.sort(key=lambda x: isinstance(x.src, Vendor))

    warehouse_names = []
    for w in warehouses:
        warehouse_names.append(get_object_or_404(Warehouse, pk=w).name)
    comp_status.sort(key=shortage_key)

    return render(request, 'epic/order_check_stock.html',
                  {
                      'breadcrumb': crumb,
                      'order': order,
                      'shortage_count': len(shortages),
                      'open_orders': open_orders,
                      'comp_status': comp_status,
                      'warehouse_names': warehouse_names,
                      'makeup_order_list': makeup_order_list,
                      'needed_by': needed_by,
                      'max_lead_time': max_lead_time,
                      'available_time': available_time,
                      'min_assy_arrival_date': min_assy_arrival_date,
                      'max_comp': max_comp,
                      'max_assy': max_assy
                  })

@permission_required(perms.EDIT)
def order_add_shipment(request, pk):
    return ship_edit(request, None, order_id=pk)

class Ship_List_View(generic.ListView):
    model = Shipment
    paginate_by = 50

    def get_queryset(self):
        return super(Ship_List_View, self).get_queryset().order_by('-id')

    def get_context_data(self, **kwargs):
        context = super(Ship_List_View, self).get_context_data(**kwargs)
        context['breadcrumb'] = breadcrumb(reverse('epic:ship_list'))
        context['list_pager'] = html_list_pager(context['page_obj'])
        return context

class Ship_Detail_View(generic.DetailView):
    model = Shipment

    def get_context_data(self, **kwargs):
        ship = self.object
        pk = ship.id
        line_items = Line_Item.objects.filter(txtn_id=pk)

        context = super(Ship_Detail_View, self).get_context_data(**kwargs)
        context['breadcrumb'] = breadcrumb(reverse('epic:ship_detail',
                                                   kwargs={'pk': pk}))
        context['page_nav'] = html_page_nav('epic:ship_detail', Shipment, pk)
        context['shipment'] = ship
        context['line_items'] = line_items
        return context

class Ship_Edit_Form(autocomplete_light.ModelForm):
    def __init__(self, fixed, *args, **kwargs):
        super(Ship_Edit_Form, self).__init__(*args, **kwargs)
        self.fixed = fixed
        self.helper = crispy_form_helper()
        self.has_assemblies = False
        self.fields['ts'].label = 'Ship Date'
        self.fields['from_warehouse'].label = 'Ship-From Warehouse'
        self.fields['warehouse'].label = 'Ship-To Warehouse'

        if self.fixed:
            if 'order' in fixed:
                self.fields['ordr'].widget.attrs['readonly'] = 'True'
            else:
                del self.fields['ordr']
            if 'from_warehouse' in fixed:
                self.fields['from_warehouse'].widget.attrs['readonly'] = 'True'
            else:
                del self.fields['from_warehouse']

    def clean(self):
        cleaned_data = super(Ship_Edit_Form, self).clean()

        if 'order' in self.fixed:
            cleaned_data['ordr'] = get_object_or_404(Order,
                                                     pk=self.fixed['order'])
        if 'from_warehouse' in self.fixed:
            cleaned_data['from_warehouse'] \
                = get_object_or_404(Warehouse, pk=self.fixed['from_warehouse'])

        dst_warehouse = cleaned_data.get('warehouse')

        order = cleaned_data.get('ordr')
        from_warehouse = cleaned_data.get('from_warehouse')

        if from_warehouse is not None:
            if order is not None:
                raise forms.ValidationError(
                    { 'ordr': ['Order # and Ship-From Warehouse '
                               'cannot both be non-empty.' ] })
            if from_warehouse == dst_warehouse:
                raise forms.ValidationError(
                    { 'warehouse': ['Must be different from '
                                    'Ship-From Warehouse.' ] })
        else:
            if order is None:
                raise forms.ValidationError(
                    { 'ordr': ['One of Order # or Ship-From Warehouse '
                               'must be specified.' ] })

    class Meta:
        model = Shipment
        fields = [ 'ts', 'ordr', 'from_warehouse', 'warehouse',
                   'tracking', 'cost_freight', 'cost_other', 'cost_discount',
                   'notes' ]
        widgets = {
            'ts': DateTimePicker(options={'format': 'YYYY-MM-DD',
                                          'pickTime': False}),
            'ordr': AutocompleteTextWidget('Order_Autocomplete')
        }

@permission_required(perms.VIEW)
def ship_list(request):
    return Ship_List_View.as_view()(request)

@permission_required(perms.VIEW)
def ship_detail(request, pk):
    return Ship_Detail_View.as_view()(request, pk=pk)

@transaction.atomic
@permission_required(perms.EDIT)
def ship_edit(request, pk, order_id=None, from_warehouse_id=None,
              ship_initial=None, items_initial=None):
    def prev_url(pk):
        if pk is None:
            if 'order' in fixed:
                return reverse('epic:order_detail',
                               kwargs={'pk': fixed['order']})
            elif 'from_warehouse' in fixed:
                return reverse('epic:warehouse_detail',
                               kwargs={'pk': fixed['from_warehouse']})
            else:
                return reverse('epic:ship_list')
        return reverse('epic:ship_detail', kwargs={'pk': pk})

    # See http://bit.ly/1wfFNKa:
    if ship_initial is None:
        ship_initial = {}
    if items_initial is None:
        items_initial = []

    if order_id is not None:
        fixed = { 'order': order_id }
    elif from_warehouse_id is not None:
        fixed = { 'from_warehouse': from_warehouse_id }
    else:
        fixed = {}

    order = None
    if pk is None:
        ship = None
        warehouse = None
        if order_id is not None:
            my_url = reverse('epic:order_add_shipment',
                             kwargs={'pk': order_id})
        elif from_warehouse_id is not None:
            my_url = reverse('epic:warehouse_add_shipment',
                             kwargs={'pk': from_warehouse_id })
        else:
            my_url = reverse('epic:ship_add')

        if 'order' in fixed:
            order = get_object_or_404(Order, pk=fixed['order'])
            if len(items_initial) == 0:
                # pre-populated with order-items that remain to be shipped:
                for item in Line_Item.objects.filter(txtn_id=fixed['order']) \
                                             .order_by('index'):
                    qty = item.qty_remaining_to_ship()
                    if qty > 0:
                        items_initial.append({ 'qty': qty,
                                               'part': item.part.id,
                                               'line_cost': item.line_cost })
        elif 'from_warehouse' in fixed:
            warehouse = int(fixed['from_warehouse'])
            if len(items_initial) == 0:
                # pre-populate with in-stock items of the warehouse:
                inv = get_stock()
                for p in inv.warehouse.get(warehouse, []):
                    qty = inv.warehouse[warehouse][p].qty
                    if qty > 0:
                        items_initial.append({ 'qty': qty, 'part': p })
    else:
        ship = get_object_or_404(Shipment, pk=pk)
        order = ship.ordr		# may be None
        warehouse = ship.from_warehouse	# may be None
        my_url = reverse('epic:ship_edit', kwargs={'pk': pk})

    widgets = {
        'part': AutocompleteTextWidget('Part_Autocomplete')
    }
    num_extra = len(items_initial) + 3

    Line_Item_Formset = inlineformset_factory(Shipment, Line_Item,
                                              formset=Line_Item_Edit_Formset,
                                              fields=[ 'part', 'qty',
                                                       'line_cost'],
                                              widgets=widgets, extra=num_extra)

    if request.method == 'POST' and 'preset' not in request.POST:
        if 'cancel' in request.POST:
            return HttpResponseRedirect(prev_url(pk))

        ship_form = Ship_Edit_Form(fixed, request.POST, instance=ship)
        item_form = Line_Item_Formset(request.POST, instance=ship)

        if ship_form.is_valid() and item_form.is_valid():
            pk = item_form.save_with_transaction(request, Line_Item_Formset,
                                                 ship_form, item_form)
            if 'save-and-done' in request.POST:
                return HttpResponseRedirect(prev_url(pk))
            return HttpResponseRedirect(reverse('epic:ship_edit',
                                                kwargs={'pk': pk}))
    else:
        if ship is None and len(ship_initial) == 0:
            ship_initial['ts'] = datetime.today()
            if order is not None:
                ship_initial['ordr'] = order.id
                ship_initial['warehouse'] = order.warehouse_id
            if warehouse is not None:
                ship_initial['from_warehouse'] = warehouse
        ship_form = Ship_Edit_Form(fixed, instance=ship, initial=ship_initial)
        item_form = Line_Item_Formset(instance=ship, initial=items_initial)

    vendor_id = None if order is None else order.vendor.id
    item_form.initialize_virtual_fields(vendor_id)

    crumb = breadcrumb(my_url)
    return render(request, 'epic/shipment_edit.html',
                  {
                      'breadcrumb': crumb,
                      'pk': pk,
                      'ship_form': ship_form,
                      'item_form': item_form,
                  })

@permission_required(perms.VIEW)
def ship_add(request):
    from_warehouse_id = None
    ship_initial = {}
    items_initial = []
    if request.method == 'POST' and 'preset' in request.POST:
        ship_initial, items_initial = get_initial_from_post(request,
                                                            'item_set',
                                                            Shipment,
                                                            Line_Item)
        from_warehouse_id = ship_initial.get('from_warehouse')
    return ship_edit(request, None, from_warehouse_id=from_warehouse_id,
                     ship_initial=ship_initial, items_initial=items_initial)

@permission_required(perms.EDIT)
def ship_delete(request, pk):
    ship = get_object_or_404(Shipment, pk=pk)
    my_url = reverse('epic:ship_delete', kwargs={'pk': pk})
    parent_url = get_parent_url(my_url)

    # Shipments are easy: nothing else refers to them so we can always delete
    # them.
    ship.delete()
    return HttpResponseRedirect(get_parent_url(parent_url))

@permission_required(perms.VIEW)
def search_results(request):
    def find_parts(q):
        # EP<number> or <number> results in direct part id lookup:
        m = re.match(r'(?:EP)?(\d+)$', q, re.I)
        if m:
            qs = Part.objects.filter(id=int(m.group(1)))
            if qs.exists():
                return qs
        qs  = Part.objects.filter(mfg_pn__icontains=q)
        qs |= Part.objects.filter(mfg__icontains=q)
        qs |= Part.objects.filter(descr__icontains=q)
        qs |= Part.objects.filter(vendor_part__vendor__name__icontains=q)
        qs |= Part.objects.filter(vendor_part__vendor_pn__icontains=q)
        qs |= Part.objects.filter(val__icontains=q)
        return qs.distinct().order_by('id')

    def find_txtns(q):
        m = re.match(r'(\d+)$', q, re.I)
        qs = Transaction.objects.none()
        if m:
            qs = Transaction.objects.filter(pk=int(m.group(1)))
            if len(qs) == 1:
                return qs
            for txtn in Transaction.objects.all():
                if re.match('\d*%s\d*' % q, str(txtn.id)):
                    qs |= Transaction.objects.filter(pk=txtn.pk)
            if len(qs) > 0:
                return qs
        qs |= Transaction.objects.filter(warehouse__name__icontains=q)
        qs |= Transaction.objects.filter(notes__icontains=q)
        qs |= Transaction.objects.filter(order__vendor__name__icontains=q)
        qs |= Transaction.objects.filter(shipment__tracking__icontains=q)
        return qs.distinct().order_by('-ts')

    parts = Part.objects.none()
    txtns = Transaction.objects.none()
    queries = [ ]
    if request.method == 'GET' and ('q' in request.GET):
        q = request.GET['q'].strip()
        queries.append('q=' + q)
        if q != '':
            parts = find_parts(q)
            txtns = find_txtns(q)

    if len(parts) == 1 and len(txtns) == 0:
        return HttpResponseRedirect(reverse('epic:part_detail',
                                            kwargs={'pk': parts[0].pk}))
    if len(txtns) == 1 and len(parts) == 0:
        txtn = txtns[0]
        if hasattr(txtn, 'order'):
            url = reverse('epic:order_detail', kwargs={'pk': txtn.pk})
        elif hasattr(txtn, 'shipment'):
            url = reverse('epic:ship_detail', kwargs={'pk': txtn.pk})
        elif hasattr(txtn, 'inventory'):
            url = reverse('epic:warehouse_inventory_detail',
                           kwargs={'warehouse': txtn.warehouse.id,
                                   'pk': txtn.pk})
        return HttpResponseRedirect(url)

    paginator = Paginator(parts, 10)
    ppg = request.GET.get('ppg')
    try:
        parts = paginator.page(ppg)
    except PageNotAnInteger:
        parts = paginator.page(1)
    except EmptyPage:
        parts = paginator.page(paginator.num_pages)

    paginator = Paginator(txtns, 10)
    tpg = request.GET.get('tpg')
    try:
        txtns = paginator.page(tpg)
    except PageNotAnInteger:
        txtns = paginator.page(1)
    except EmptyPage:
        txtns = paginator.page(paginator.num_pages)

    my_url = reverse('epic:search_results')
    crumb = breadcrumb(my_url)
    return render(request, 'epic/search_results.html',
                  {
                      'q':		q,
                      'breadcrumb':	breadcrumb,
                      'parts':		parts,
                      'parts_pager':	html_list_pager(parts,
                                                        queries=queries,
                                                        key='ppg'),
                      'txtns':		txtns,
                      'txtns_pager':	html_list_pager(txtns,
                                                        queries=queries,
                                                        key='tpg')
                  })

@permission_required(perms.VIEW)
def epic_index(request):
    open_order_summaries = get_open_order_summary()
    inv = get_stock()
    orders_with_shortages = []
    for summary in open_order_summaries:
        # update current stock:
        for line_item in summary['items']:
            inv.apply_order_line_item(line_item)

        try:
            warehouse = Warehouse.by_name(summary['order'].vendor)
        except:
            continue

        w = warehouse.id

        has_shortages = 0
        for line_item in summary['items']:
            for assy_item in line_item.part.assembly_items():
                part = assy_item.comp.best_part()
                qty = 0
                for p in part.equivalent_parts():
                    qty += inv.get_qty(w, p.id)
                if qty < 0:
                    orders_with_shortages.append(
                        summary['order'].transaction_ptr)
                    has_shortages = True
                    break
            if has_shortages:
                break

    orders_with_shortages.sort(key=operator.attrgetter('ts'), reverse=True)

    now = timezone.now()
    ts_limit = now - timedelta(days=30)

    activity = []

    recent = Transaction.objects.filter(ts__gte=ts_limit) \
                                .order_by('-ts')[:20]
    for txtn in recent:
        try:
            orders_with_shortages.index(txtn)
        except:
            activity.append(txtn)

    past_due = Order.objects \
                    .filter(status=Order.STATUS_OPEN) \
                    .filter(expected_arrival_date__lte=now) \
                    .order_by('-ts')
    for txtn in past_due:
        try:
            activity.index(txtn)
        except:
            activity.append(txtn)

    breadcrumb = '<ol class="breadcrumb">' \
                 '%s: %s</ol>' % (EPIC_App_Config.name.upper(),
                                  EPIC_App_Config.verbose_name)
    return render(request, 'epic/epic_index.html',
                  {
                      'breadcrumb': breadcrumb,
                      'orders_with_shortages': orders_with_shortages,
                      'activity': activity
                  })



#
#   Copyright (c) 2014-2015 eGauge Systems LLC
# 	4730 Walnut St, Suite 110
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
from django.apps import AppConfig

class EPIC_App_Config(AppConfig):
    name = 'epic'
    verbose_name = 'Electronic Parts Inventory Center'

    def ready(self):
        import epic.signals
